from __future__ import annotations

from collections.abc import Sequence

import numpy as np

from ._shared import ArrayType


class DataArray(Sequence):
    """Python wrapper for .NET DataArray class.

    Parameters
    ----------
    psarray
        Reference to .NET DataArray object.
    """

    def __init__(self, *, psarray):
        self._psarray = psarray

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'name={self.name}, '
            f'unit={self.unit}, '
            f'n_points={len(self)})'
        )

    def __getitem__(self, index):
        if isinstance(index, int):
            if index >= len(self) or index < -len(self):
                raise IndexError('list index out of range')
            index = index % len(self)
            return self._psarray[index].Value

        return self.to_list()[index]

    def __len__(self):
        return len(self._psarray)

    def copy(self) -> DataArray:
        """Return a copy of the array."""
        return DataArray(psarray=self._psarray.Clone())

    def min(self) -> float:
        """Return min value."""
        return self._psarray.MinValue

    def max(self) -> float:
        """Return max value."""
        return self._psarray.MaxValue

    def savitsky_golay(self, window_size: int = 3) -> DataArray:
        """Smooth the array using a Savitsky-Golay filter with the window size.

        (i.e. window size 2 will filter points based on the values of the next/previous 2 points)

        Parameters
        ----------
        window_size : int
            Size of the window
        """
        new = self.copy()
        success = new._psarray.Smooth(window_size, False)
        if not success:
            raise ValueError('Something went wrong.')
        return new

    @property
    def name(self) -> str:
        """Name of the array."""
        return self._psarray.Description

    def to_numpy(self) -> np.ndarray:
        """Export data array to numpy."""
        return np.array(self._psarray.GetValues())

    def to_list(self) -> list[float]:
        """Export data array to list."""
        return list(self._psarray.GetValues())

    @property
    def type(self) -> ArrayType:
        """ArrayType enum."""
        return ArrayType(self._psarray.ArrayType)

    @property
    def unit(self) -> str:
        """Unit for array."""
        return self._psarray.Unit.ToString()

    @property
    def quantity(self) -> str:
        """Quantity for array."""
        return self._psarray.Unit.Quantity

    @property
    def ocp_value(self) -> float:
        """OCP Value."""
        return self._psarray.OCPValue



from __future__ import annotations

from dataclasses import dataclass

from PalmSens import Measurement as PSMeasurement

from .._fitting import FitResult
from .._methods.method import Method
from .curve import Curve
from .dataset import DataSet
from .eisdata import EISData
from .peak import Peak


@dataclass(frozen=True)
class DeviceInfo:
    """Dataclass for device information."""

    type: str
    """Device type"""
    firmware: str
    """Firmware version"""
    serial: str
    """Serial number"""
    id: int
    """Device ID"""

    @classmethod
    def _from_psmeasurement(cls, obj: PSMeasurement) -> DeviceInfo:
        """Construct device dataclass from SDK measurement object."""
        return cls(
            type=obj.DeviceUsed.ToString(),
            firmware=obj.DeviceUsedFW,
            serial=obj.DeviceUsedSerial,
            id=int(obj.DeviceUsed),
        )


@dataclass
class Measurement:
    """Python wrapper for .NET Measurement class.

    Parameters
    ----------
    psmeasurement
        Reference to .NET measurement object.
    """

    def __init__(self, *, psmeasurement: PSMeasurement):
        self._psmeasurement = psmeasurement

    def __repr__(self):
        return f'{self.__class__.__name__}(title={self.title}, timestamp={self.timestamp}, device={self.device.type})'

    @property
    def title(self) -> str:
        """Title for the measurement."""
        return self._psmeasurement.Title

    @property
    def timestamp(self) -> str:
        """Date and time of the start of this measurement.."""
        return str(self._psmeasurement.TimeStamp)

    @property
    def device(self) -> DeviceInfo:
        """Return dataclass with measurement device information."""
        return DeviceInfo._from_psmeasurement(self._psmeasurement)

    @property
    def blank_curve(self) -> Curve | None:
        """Blank curve.

        if Blank curve is present (not null) a new curve will be added after each measurement
        containing the result of the measured curve subtracted with the Blank curve.
        """
        curve = self._psmeasurement.BlankCurve
        if curve:
            return Curve(pscurve=curve)
        return None

    @property
    def has_blank_subtracted_curves(self) -> bool:
        """Return True if the curve collection contains a blank subtracted curve."""
        return self._psmeasurement.ContainsBlankSubtractedCurves

    @property
    def has_eis_data(self) -> bool:
        """Return True if EIS data are is available."""
        return self._psmeasurement.ContainsEISData

    @property
    def dataset(self) -> DataSet:
        """Dataset containing multiple arrays of values.

        All values are related by means of their indices.
        Data arrays in a dataset should always have an equal amount of entries.
        """
        return DataSet(psdataset=self._psmeasurement.DataSet)

    @property
    def eis_data(self) -> list[EISData]:
        """EIS data in measurement."""
        lst = [EISData(pseis=pseis) for pseis in self._psmeasurement.EISdata]

        return lst

    @property
    def method(self) -> Method:
        """Method related with this Measurement.

        The information from the Method is used when saving Curves."""
        return Method(psmethod=self._psmeasurement.Method)

    @property
    def channel(self) -> float:
        """Get the channel that the measurement was measured on."""
        return self._psmeasurement.Channel

    @property
    def ocp_value(self) -> float:
        """First OCP Value from either curves or EISData."""
        return self._psmeasurement.OcpValue

    @property
    def n_curves(self) -> int:
        """Number of curves that are part of the Measurement class."""
        return self._psmeasurement.nCurves

    @property
    def n_eis_data(self) -> int:
        """Number of EISdata curves (channels) that are part of the Measurement class."""
        return self._psmeasurement.nEISdata

    @property
    def peaks(self) -> list[Peak]:
        """Get peaks from all curves.

        Returns
        -------
        peaks : list[Peak]
            List of peaks
        """
        peaks = []
        for curve in self.curves:
            peaks.extend(curve.peaks)
        return peaks

    @property
    def eis_fit(self) -> list[FitResult]:
        """Get all EIS fits from measurement

        Returns
        -------
        eis_fits : list[EISFitResults]
            Return list of EIS fits
        """
        eisdatas = self.eis_data
        eis_fits = [FitResult.from_eisdata(eisdata) for eisdata in eisdatas]
        return eis_fits

    @property
    def curves(self) -> list[Curve]:
        """Get all curves in measurement.

        Returns
        -------
        curves : list[Curve]
            List of curves
        """
        curves = self._psmeasurement.GetCurveArray()
        return [Curve(pscurve=curve) for curve in curves]



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal, Sequence

import PalmSens

from .._shared import single_to_double


class CURRENT_RANGE(Enum):
    """Get the id for a given current range.

    Use these when defining a current range."""

    cr_100_pA = 0
    """100 pA"""
    cr_1_nA = 1
    """1 nA"""
    cr_10_nA = 2
    """10 nA"""
    cr_100_nA = 3
    """100 nA"""
    cr_1_uA = 4
    """1 μA"""
    cr_10_uA = 5
    """10 μA"""
    cr_100_uA = 6
    """100 μA"""
    cr_1_mA = 7
    """1 mA"""
    cr_10_mA = 8
    """10 mA"""
    cr_100_mA = 9
    """100 mA"""
    cr_2_uA = 10
    """2 μA"""
    cr_4_uA = 11
    """4 μA"""
    cr_8_uA = 12
    """8 μA"""
    cr_16_uA = 13
    """16 μA"""
    cr_32_uA = 14
    """32 μA"""
    cr_63_uA = 26
    """63 μA"""
    cr_125_uA = 17
    """125 μA"""
    cr_250_uA = 18
    """250 μA"""
    cr_500_uA = 19
    """500 μA"""
    cr_5_mA = 20
    """5 mA"""
    cr_6_uA = 21
    """6 μA"""
    cr_13_uA = 22
    """13 μA"""
    cr_25_uA = 23
    """25 μA"""
    cr_50_uA = 24
    """50 μA"""
    cr_200_uA = 25
    """200 μA"""
    cr_1_A = 30
    """1 A"""

    def _to_psobj(self):
        """Get equivalent PS object."""
        return PalmSens.CurrentRange(PalmSens.CurrentRanges(self.value))

    @classmethod
    def _from_psobj(cls, psobj, /):
        """Convert from PS object."""
        return cls(int(PalmSens.CurrentRange.GetCRfromCRByte(psobj.CRbyte)))


class POTENTIAL_RANGE(Enum):
    """Get the id for a given current range.

    Use these when defining a potential range."""

    pr_1_mV = 0
    """1 mV"""
    pr_10_mV = 1
    """10 mV"""
    pr_20_mV = 2
    """20 mV"""
    pr_50_mV = 3
    """50 mV"""
    pr_100_mV = 4
    """100 mV"""
    pr_200_mV = 5
    """200 mV"""
    pr_500_mV = 6
    """500 mV"""
    pr_1_V = 7
    """1 V"""

    def _to_psobj(self):
        """Get equivalent PS object."""
        return PalmSens.PotentialRange(PalmSens.PotentialRanges(self.value))

    @classmethod
    def _from_psobj(cls, psobj, /):
        """Convert from PS object."""
        return cls(int(PalmSens.PotentialRange.get_PR(psobj)))


def convert_bools_to_int(lst: Sequence[bool]) -> int:
    """Convert e.g. [True, False, True, False] to 5."""
    return int(''.join('01'[set_high] for set_high in reversed(lst)), base=2)


def convert_int_to_bools(val) -> tuple[bool, bool, bool, bool]:
    """Convert e.g. 5 to [True, False, True, False]."""
    lst = tuple([bool(int(_)) for _ in reversed(f'{val:04b}')])
    assert len(lst) == 4  # specify length to make mypy happy
    return lst


def set_extra_value_mask(
    obj,
    *,
    enable_bipot_current: bool = False,
    record_auxiliary_input: bool = False,
    record_cell_potential: bool = False,
    record_dc_current: bool = False,
    record_we_potential: bool = False,
    record_forward_and_reverse_currents: bool = False,
    record_we_current: bool = False,
):
    """Set the extra value mask for a given method."""
    extra_values = 0

    for flag, enum in (
        (enable_bipot_current, PalmSens.ExtraValueMask.BipotWE),
        (record_auxiliary_input, PalmSens.ExtraValueMask.AuxInput),
        (record_cell_potential, PalmSens.ExtraValueMask.CEPotential),
        (record_dc_current, PalmSens.ExtraValueMask.DCcurrent),
        (record_we_potential, PalmSens.ExtraValueMask.PotentialExtraRE),
        (record_forward_and_reverse_currents, PalmSens.ExtraValueMask.IForwardReverse),
        (record_we_current, PalmSens.ExtraValueMask.CurrentExtraWE),
    ):
        if flag:
            extra_values = extra_values | int(enum)

    obj.ExtraValueMsk = PalmSens.ExtraValueMask(extra_values)


def get_extra_value_mask(obj) -> dict[str, Any]:
    mask = obj.ExtraValueMsk

    ret = {
        'enable_bipot_current': mask.HasFlag(PalmSens.ExtraValueMask.BipotWE),
        'record_auxiliary_input': mask.HasFlag(PalmSens.ExtraValueMask.AuxInput),
        'record_cell_potential': mask.HasFlag(PalmSens.ExtraValueMask.CEPotential),
        'record_dc_current': mask.HasFlag(PalmSens.ExtraValueMask.DCcurrent),
        'record_we_potential': mask.HasFlag(PalmSens.ExtraValueMask.PotentialExtraRE),
        'record_forward_and_reverse_currents': mask.HasFlag(
            PalmSens.ExtraValueMask.IForwardReverse
        ),
        'record_we_current': mask.HasFlag(PalmSens.ExtraValueMask.CurrentExtraWE),
    }

    return ret


@dataclass
class ELevel:
    """Create a multi-step amperometry level method object."""

    level: float = 0.0
    """Level in V."""

    duration: float = 1.0
    """Duration in s."""

    record: bool = True
    """Record the current."""

    limit_current_max: float | None = None
    """Limit current max in µA. Set to None to disable."""

    limit_current_min: float | None = None
    """Limit current min in µA. Set to None to disable."""

    trigger_lines: Sequence[Literal[0, 1, 2, 3]] = field(default_factory=list)
    """Trigger at level lines.

    Set digital output lines at start of measurement, end of equilibration.
    Accepted values: 0 for d0, 1 for d1, 2 for d2, 3 for d3.
    """

    @property
    def use_limits(self) -> bool:
        """Return True if instance sets current limits."""
        use_limit_current_min = self.limit_current_min is not None
        use_limit_current_max = self.limit_current_max is not None

        return use_limit_current_min or use_limit_current_max

    def to_psobj(self):
        obj = PalmSens.Techniques.ELevel()

        obj.Level = self.level
        obj.Duration = self.duration
        obj.Record = self.record

        obj.UseMaxLimit = self.limit_current_max is not None
        obj.MaxLimit = self.limit_current_max or 0.0
        obj.UseMinLimit = self.limit_current_min is not None
        obj.MinLimit = self.limit_current_min or 0.0

        obj.UseTriggerOnStart = bool(self.trigger_lines)

        trigger_bools = [(val in self.trigger_lines) for val in (0, 1, 2, 3)]

        obj.TriggerValueOnStart = convert_bools_to_int(trigger_bools)

        return obj

    @classmethod
    def from_psobj(cls, psobj: PalmSens.Techniques.ELevel):
        """Construct ELevel dataclass from PalmSens.Techniques.ELevel object."""
        trigger_lines: list[Literal[0, 1, 2, 3]] = []

        if psobj.UseTriggerOnStart:
            trigger_bools = convert_int_to_bools(psobj.TriggerValueOnStart)
            for i in (0, 1, 2, 3):
                if trigger_bools[i]:
                    trigger_lines.append(i)

        return cls(
            level=single_to_double(psobj.Level),
            duration=single_to_double(psobj.Duration),
            record=psobj.Record,
            limit_current_max=single_to_double(psobj.MaxLimit) if psobj.MaxLimit else None,
            limit_current_min=single_to_double(psobj.MinLimit) if psobj.MinLimit else None,
            trigger_lines=trigger_lines,
        )


@dataclass
class ILevel:
    """Create a multi-step potentiometry level method object."""

    level: float = 0.0
    """Level in I.

    This value is multiplied by the applied current range."""

    duration: float = 1.0
    """Duration in s."""

    record: bool = True
    """Record the current."""

    limit_potential_max: float | None = None
    """Limit potential max in V. Set to None to disable."""

    limit_potential_min: float | None = None
    """Limit potential min in V. Set to None to disable."""

    trigger_lines: Sequence[Literal[0, 1, 2, 3]] = field(default_factory=list)
    """Trigger at level lines.

    Set digital output lines at start of measurement, end of equilibration.
    Accepted values: 0 for d0, 1 for d1, 2 for d2, 3 for d3.
    """

    @property
    def use_limits(self) -> bool:
        """Return True if instance sets current limits."""
        use_limit_potential_min = self.limit_potential_min is not None
        use_limit_potential_max = self.limit_potential_max is not None

        return use_limit_potential_min or use_limit_potential_max

    def to_psobj(self):
        obj = PalmSens.Techniques.ELevel()

        obj.Level = self.level
        obj.Duration = self.duration
        obj.Record = self.record

        obj.UseMaxLimit = self.limit_potential_max is not None
        obj.MaxLimit = self.limit_potential_max or 0.0
        obj.UseMinLimit = self.limit_potential_min is not None
        obj.MinLimit = self.limit_potential_min or 0.0

        obj.UseTriggerOnStart = bool(self.trigger_lines)

        trigger_bools = [(val in self.trigger_lines) for val in (0, 1, 2, 3)]

        obj.TriggerValueOnStart = convert_bools_to_int(trigger_bools)

        return obj

    @classmethod
    def from_psobj(cls, psobj: PalmSens.Techniques.ELevel):
        """Construct ILevel dataclass from PalmSens.Techniques.ELevel object."""
        trigger_lines: list[Literal[0, 1, 2, 3]] = []

        if psobj.UseTriggerOnStart:
            trigger_bools = convert_int_to_bools(psobj.TriggerValueOnStart)
            for i in (0, 1, 2, 3):
                if trigger_bools[i]:
                    trigger_lines.append(i)

        return cls(
            level=single_to_double(psobj.Level),
            duration=single_to_double(psobj.Duration),
            record=psobj.Record,
            limit_potential_max=single_to_double(psobj.MaxLimit) if psobj.MaxLimit else None,
            limit_potential_min=single_to_double(psobj.MinLimit) if psobj.MinLimit else None,
            trigger_lines=trigger_lines,
        )



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Optional

from PalmSens import Fitting as PSFitting
from System import Array

from ._data.curve import Curve
from ._data.eisdata import EISData

if TYPE_CHECKING:
    from matplotlib import figure


@dataclass(slots=True)
class Parameter:
    """Set or update Parameter attributes."""

    symbol: str
    """Name of the parameter (not used in minimization)."""
    value: Optional[float] = None
    """Initial value of the parameter."""
    min: Optional[float] = None
    """Minimum (lower bound) for the parameter."""
    max: Optional[float] = None
    """Maximum (upper bound) for the parameter."""
    fixed: Optional[bool] = None
    """If True, fix the value for this parameter."""

    @classmethod
    def _from_psparameter(cls, psparameter: PSFitting.Parameter):
        """Create instance from SDK Parameter object."""
        return cls(
            symbol=psparameter.Symbol,
            value=psparameter.Value,
            min=psparameter.MinValue,
            max=psparameter.MaxValue,
            fixed=psparameter.Fixed,
        )

    def _update_psparameter(self, psparameter: PSFitting.Parameter):
        """Update PalmSens SDK object with values from dataclass."""
        if self.value:
            psparameter.Value = self.value
        if self.min:
            psparameter.MinValue = self.min
        if self.max:
            psparameter.MaxValue = self.max
        if self.fixed:
            psparameter.Fixed = self.fixed


class Parameters(Sequence):
    """Tuple-like container class for parameters.

    This class is instantiated from the CDC code and contains
    default parameters. This ensures that the length and type of
    parameters match that of `CircuitModel`. Update the parameters
    in this class and pass to `CircuitModel.fit()`.

    Parameters
    ----------
    cdc: str
        Genererate fitting parameters for this CDC.
    """

    def __init__(self, cdc: str):
        self.cdc = cdc
        """CDC code used to generate parameter listing."""

        model = PSFitting.Models.CircuitModel()
        model.SetCircuit(cdc)
        self._parameters = tuple(
            Parameter._from_psparameter(psparam) for psparam in model.InitialParameters
        )

    def __len__(self):
        return len(self._parameters)

    def __getitem__(self, key):
        return self._parameters[key]

    def __repr__(self) -> str:
        return self._parameters.__repr__()

    def __str__(self) -> str:
        return self._parameters.__str__()

    def _update_psmodel_parameters(self, psmodel: PSFitting.Models.CircuitModel) -> None:
        """Update the initial parameters in the SDK model with parameters in this instance.

        Note that the length and type of parameters must match that of the SDK class.
        """
        if len(self) != psmodel.NParameters:
            raise ValueError(f'Parameters must be of length {psmodel.NParameters}')

        for param, psparam in zip(self, psmodel.InitialParameters):
            param._update_psparameter(psparam)


@dataclass(frozen=True)
class FitResult:
    """Container for fitting results."""

    cdc: str
    """Circuit model CDC values."""
    parameters: list[float]
    """Optimized parameters for CDC."""
    error: list[float]
    """Error (%) on parameters."""
    chisq: float
    """Chi-squared goodness of fit statistic."""
    n_iter: int
    """Total number of iterations."""
    exit_code: str
    """Exit code for the minimization."""

    @classmethod
    def from_psfitresult(cls, result: PSFitting.FitResult, cdc: str):
        """Construct fitresult from SDK FitResult."""
        return cls(
            cdc=cdc,
            chisq=result.ChiSq,
            exit_code=result.ExitCode.ToString(),
            n_iter=result.NIterations - 1,
            parameters=list(result.FinalParameters),
            error=list(result.ParameterSDs),
        )

    @classmethod
    def from_eisdata(cls, data: EISData):
        """Construct fitresulf from EISData."""
        return cls(
            cdc=data.cdc,
            parameters=data.cdc_values,
            chisq=0,
            error=[0.0 for _ in data.cdc_values],
            exit_code='',
            n_iter=0,
        )

    def get_psmodel(self, data: EISData) -> PSFitting.Models.CircuitModel:
        """Get SDK Circuit model object"""
        psmodel = PSFitting.Models.CircuitModel()
        psmodel.SetEISdata(data._pseis)
        psmodel.SetCircuit(self.cdc)
        psmodel.SetInitialParameters(self.parameters)
        return psmodel

    def get_nyquist(self, data: EISData) -> tuple[Curve, Curve]:
        """Calculate observed and calculated nyquist curves.

        Parameters
        ----------
        data : EISData
            Input EIS data.

        Returns
        -------
        calc, meas : tuple[Curve, Curve]
            Returns the nyquist curve calculated from the model parameters
            and the measured curve from the EIS data.
        """
        psmodel = self.get_psmodel(data=data)
        curves = psmodel.GetNyquist()
        calc, meas = (Curve(pscurve=pscurve) for pscurve in curves)
        return calc, meas

    def get_bode_z(self, data: EISData) -> tuple[Curve, Curve]:
        """Calculate observed and calculated Bode curve Z vs Frequency.

        Parameters
        ----------
        data : EISData
            Input EIS data.

        Returns
        -------
        calc, meas : tuple[Curve, Curve]
            Returns the nyquist curve calculated from the model parameters
            and the measured curve from the EIS data.
        """
        psmodel = self.get_psmodel(data=data)
        curves = psmodel.GetCurveZabsOverFrequency(False)
        calc, meas = (Curve(pscurve=pscurve) for pscurve in curves)
        return calc, meas

    def get_bode_phase(self, data: EISData) -> tuple[Curve, Curve]:
        """Calculate observed and calculated Bode curve phase vs Frequency.

        Parameters
        ----------
        data : EISData
            Input EIS data.

        Returns
        -------
        calc, meas : tuple[Curve, Curve]
            Returns the nyquist curve calculated from the model parameters
            and the measured curve from the EIS data.
        """
        psmodel = self.get_psmodel(data=data)
        curves = psmodel.GetCurvePhaseOverFrequency(False)
        calc, meas = (Curve(pscurve=pscurve) for pscurve in curves)
        return calc, meas

    def plot_nyquist(self, data: EISData) -> figure.Figure:
        """Make nyquist plot using matplotlib.

        Parameters
        ----------
        data : EISData
            Input EIS data.

        Returns
        -------
        fig : fig.Figure
            Returns matplotlib figure object. use `fig.show()` to render plot.
        """
        import matplotlib.pyplot as plt

        calc, meas = self.get_nyquist(data=data)
        fig, ax = plt.subplots()
        ax.set_title('Nyquist plot')

        calc.plot(ax=ax)
        meas.plot(ax=ax, marker='^', linestyle='None')

        return fig

    def plot_bode(self, data: EISData) -> figure.Figure:
        """Make bode plot using matplotlib.

        Parameters
        ----------
        data : EISData
            Input EIS data.

        Returns
        -------
        fig : fig.Figure
            Returns matplotlib figure object. use `fig.show()` to render plot.
        """
        import matplotlib.pyplot as plt

        calc_z, meas_z = self.get_bode_z(data=data)
        calc_ph, meas_ph = self.get_bode_phase(data=data)
        fig, ax1 = plt.subplots()
        ax1.set_title('Bode plot')
        ax1.set_xscale('log')

        calc_z.plot(ax=ax1, legend=False, color='C0')
        meas_z.plot(ax=ax1, marker='^', linestyle='None', color='C0', legend=False)

        ax2 = ax1.twinx()
        calc_ph.plot(ax=ax2, legend=False, color='C1')
        meas_ph.plot(ax=ax2, marker='^', linestyle='None', color='C1', legend=False)

        fig.legend()
        return fig


@dataclass
class CircuitModel:
    """Fit an equivalent circuit model.

    The class takes a CDC string as a required argument to set up the model.

    The other parameters are optional and can be used to tweak the minimization.
    The model supports fitting over a specified frequency range and adjustment of exit
    conditions (i.e. max # iterations, min delta error, min parameter step
    size).

    Optionally you can change the initial values of the parameters, their
    min/max bounds or fix their value.

    Example:

    ```
    model = CircuitModel('R(RC)')
    result = model.fit(eis_data)
    ```
    """

    cdc: str
    """Sets the circuit specified in the CDC string.

    For more information, see:
        https://www.utwente.nl/en/tnw/ims/publications/downloads/cdc-explained.pdf
    """

    algorithm: Literal['leastsq', 'nelder-mead'] = 'leastsq'
    """Name of the fitting method to use.

    Valid values are: `leastsq` (Levenberg-Marquardt), `nelder-mead`
    """

    max_iterations: int = 500
    """Maximum number of iterations.

    Minimization terminates once it reaches this number of steps (default = 500).
    """

    min_delta_error: float = 1.0e-9
    """Minimum convergence error.

    Minimization converges if the residual (squared difference)
        falls below this value (default = 1e-9).
    """

    min_delta_step: float = 1.0e-12
    """Minimum convergence step.

    Minimization converges if the difference in parameter values
        falls below this value (default = 1e-12)."""

    min_freq: Optional[float] = None
    """Minimum fitting frequency in Hz."""

    max_freq: Optional[float] = None
    """Maximum fitting frequency in Hz."""

    tolerance: float = 1e-4
    """Convergence tolerance. Nelder-Mead only (default = 1e-4)."""

    lambda_start: float = 0.01
    """Start lambda value. Levenberg-Marquardt only (default = 0.01)."""

    lambda_factor: float = 10.00
    """Lambda Scaling Factor. Levenberg-Marquardt only (default = 10)."""

    _last_result: Optional[FitResult] = field(default=None, repr=False)
    _last_psfitter: Optional[PSFitting.FitAlgorithm] = field(default=None, repr=False)

    def default_parameters(self) -> Parameters:
        """Get default parameters. Use this to modify parameter values.

        Returns
        -------
        parameters : Parameters
            Default parameters for CDC.
        """
        return Parameters(self.cdc)

    def _psfitoptions(
        self,
        data: EISData,
        *,
        parameters: Optional[Sequence[float] | Parameters] = None,
    ) -> PSFitting.FitOptions:
        """Fit circuit model.

        Parameters
        ----------
        data : EISData
            Input EIS data.
        parameters : Optional[Sequence[float] | Parameters]
            Optional initial parameters for fit. Can be passed as
            `Parameters` object or list of values.

        Returns
        -------
        opts : PSFitting.FitOptions
            SDK object containing fitting options.
        """
        model = PSFitting.Models.CircuitModel()
        model.SetCircuit(self.cdc)
        model.SetEISdata(data._pseis)

        if parameters:
            if isinstance(parameters, Parameters):
                if self.cdc != parameters.cdc:
                    raise ValueError(
                        f'Parameters cdc ({self.cdc}) does not match Model ({parameters.cdc})'
                    )
                parameters._update_psmodel_parameters(model)
            else:
                if len(parameters) != model.NParameters:
                    raise ValueError(f'Parameters must be of length {model.NParameters}')
                model.SetInitialParameters(parameters)

        opts = PSFitting.FitOptionsCircuit()
        opts.Model = model
        opts.RawData = data._pseis

        opts.MaxIterations = self.max_iterations
        opts.MinimumDeltaErrorTerm = self.min_delta_error
        opts.MinimumDeltaParameters = self.min_delta_step

        if self.min_freq or self.max_freq:
            self.min_freq = self.min_freq or 0
            self.max_freq = self.max_freq or 0

            array = data.dataset.freq_arrays()[-1]
            sel = (self.min_freq < val < self.max_freq for val in array)

            opts.SelectedDataPoints = Array[bool]((bool(_) for _ in sel))

        return opts

    @property
    def last_result(self):
        """Store last fit result."""
        return self._last_result

    @property
    def last_psfitter(self):
        """Store reference to last SDK fitting object."""
        return self._last_psfitter

    def fit(
        self, data: EISData, *, parameters: Optional[Sequence[float] | Parameters] = None
    ) -> FitResult:
        """Fit circuit model.

        Parameters
        ----------
        data : EISData
            Input data.
        parameters : Optional[Sequence[float] | Parameters]
            Optional initial parameters for fit. Can be passed as
            `Parameters` object or list of values.

        Returns
        -------
        result : FitResult
            Returns dataclass with fit results. Can also be accessed via `.last_result`.
        """
        if not data.frequency_type == 'Scan':
            raise ValueError(
                f'Fit only supports EIS scans at a fixed potential, got {data.frequency_type=}.'
            )
        if not data.scan_type == 'Fixed':
            raise ValueError(
                f'Fit only supports EIS scans at a fixed potential, got {data.scan_type=}.'
            )

        opts = self._psfitoptions(data=data, parameters=parameters)

        fitter = PSFitting.FitAlgorithm.FromAlgorithm(opts)
        fitter.ApplyFitCircuit()
        self._last_psfitter = fitter
        self._last_result = FitResult.from_psfitresult(fitter.FitResult, cdc=self.cdc)
        return self._last_result



"""This module contains the public api for classes for method configuration."""

from __future__ import annotations

from ._methods._shared import (
    CURRENT_RANGE,
    POTENTIAL_RANGE,
    ELevel,
    ILevel,
)
from ._methods.settings import (
    BiPot,
    ChargeLimits,
    CurrentLimits,
    CurrentRange,
    DataProcessing,
    DelayTriggers,
    EquilibrationTriggers,
    General,
    IrDropCompensation,
    MeasurementTriggers,
    Multiplexer,
    PostMeasurement,
    PotentialLimits,
    PotentialRange,
    Pretreatment,
    VersusOCP,
)

__all__ = [
    'BiPot',
    'ChargeLimits',
    'CURRENT_RANGE',
    'CurrentLimits',
    'CurrentRange',
    'DataProcessing',
    'DelayTriggers',
    'ELevel',
    'ILevel',
    'EquilibrationTriggers',
    'General',
    'IrDropCompensation',
    'MeasurementTriggers',
    'Multiplexer',
    'PostMeasurement',
    'POTENTIAL_RANGE',
    'PotentialLimits',
    'PotentialRange',
    'Pretreatment',
    'VersusOCP',
]



from __future__ import annotations

import atexit
from importlib.resources import files
from pathlib import Path

import clr
from pythonnet import unload

PSSDK_DIR = files('pypalmsens._pssdk.win')

core_dll = PSSDK_DIR / 'PalmSens.Core.dll'
ble_dll = PSSDK_DIR / 'PalmSens.Core.Windows.BLE.dll'


def unblock(path: Path):
    """Unblock DLL: https://stackoverflow.com/q/20886450"""
    zone_id = path.with_name(path.name + ':Zone.Identifier')
    if zone_id.exists():
        zone_id.unlink()


for dll in (core_dll, ble_dll):
    assert isinstance(dll, Path)
    unblock(dll)

# This dll contains the classes in which the data is stored
clr.AddReference(str(core_dll))

# This dll is used to load your session file
clr.AddReference(str(ble_dll))

clr.AddReference('System')

from PalmSens.Windows import CoreDependencies  # noqa: E402
from System import Diagnostics  # noqa: E402

CoreDependencies.Init()

sdk_version = Diagnostics.FileVersionInfo.GetVersionInfo(str(core_dll)).ProductVersion

atexit.register(unload)



"""This module contains the public api for circuit fitting."""

from __future__ import annotations

from ._fitting import (
    CircuitModel,
    FitResult,
    Parameter,
    Parameters,
)

__all__ = [
    'CircuitModel',
    'FitResult',
    'Parameter',
    'Parameters',
]



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Callable, Generator

import clr
from PalmSens.Data import CurrentReading
from PalmSens.Plottables import Curve as PSCurve

from ._shared import ArrayType
from .curve import Curve
from .data_array import DataArray

if TYPE_CHECKING:
    import pandas as pd
    from PalmSens.Data import DataSet as PSDataSet


def _dataset_to_mapping_with_unique_keys(psdataset: PSDataSet, /) -> dict[str, DataArray]:
    """Suffix non-unique keys with integer. Keys are derived from the array type."""
    arrays = [array for array in psdataset.GetDataArrays()]
    array_types = [ArrayType(array.ArrayType).name for array in arrays]

    mapping = {}

    for array in arrays:
        array_type = ArrayType(array.ArrayType).name

        is_unique = array_types.count(array_type) == 1

        if not is_unique:
            i = 1
            while (key := f'{array_type}_{i}') in mapping:
                i += 1
        else:
            key = array_type

        mapping[key] = DataArray(psarray=array)

    return mapping


class DataSet(Mapping):
    """Python wrapper for .NET DataSet class.

    Parameters
    ----------
    psdataset : PalmSens.Data.DataSet
        Reference to .NET DataSet object.
    """

    def __init__(self, *, psdataset: PSDataSet):
        self._psdataset = psdataset
        self._mapping = _dataset_to_mapping_with_unique_keys(psdataset)

    def __repr__(self):
        return f'{self.__class__.__name__}({list(self.keys())})'

    def __getitem__(self, key: str):
        return self._mapping[key]

    def __iter__(self) -> Generator[str, None, None]:
        # Note that iterating over self.psdataset also returns the 'hidden' debug arrays
        # `.GetDataArrays()` excludes those.
        yield from self._mapping

    def __len__(self):
        return len(self._mapping)

    def _filter(self, key: Callable) -> list[DataArray]:
        """Filter array list based on callable.

        Callable takes dotnet DataArray as its only argument.
        """
        return [array for array in self._mapping.values() if key(array)]

    def _psarrays(self):
        """Return underlying PalmSens SDK objects."""
        return self._psdataset.GetDataArrays()

    def curve(self, x: str, y: str, title: str | None = None) -> Curve:
        """Construct a custom curve from x and y keys.

        Parameters
        ----------
        x : str
            Key identifying the x array
        y : str
            Key identifying the y array
        title : str
            Set the title. If None, use the $x-$y as title

        Returns
        -------
        curve : Curve
            New Curve with plotting x against y
        """
        xarray = self[x]
        yarray = self[y]

        if not title:
            title = f'{x}-{y}'

        pscurve = PSCurve(xarray._psarray, yarray._psarray, title=title)

        return Curve(pscurve=pscurve)

    def arrays(self) -> list[DataArray]:
        """Return list of all arrays. Alias for `.to_list()`"""
        return list(self.values())

    def hidden_arrays(self) -> list[DataArray]:
        """Return 'hidden' arrays used for debugging."""
        return [DataArray(psarray=psarray) for psarray in self._psdataset if psarray.Hidden]

    def arrays_by_name(self, name: str) -> list[DataArray]:
        """Get arrays by name.

        Parameters
        ----------
        name : str
            Name of the array.

        Returns
        -------
        arrays : list[DataArray]
        """
        return self._filter(key=lambda array: array.name == name)

    def arrays_by_quantity(self, quantity: str) -> list[DataArray]:
        """Get arrays by quantity.

        Parameters
        ----------
        quantity : str
            Quantity of the array.

        Returns
        -------
        arrays : list[DataArray]
        """
        return self._filter(key=lambda array: array.quantity == quantity)

    def arrays_by_type(self, array_type: ArrayType) -> list[DataArray]:
        """Get arrays by data type.

        Parameters
        ----------
        array_type : str
            Type of the array.

        Returns
        -------
        arrays : list[DataArray]
        """
        return self._filter(key=lambda array: array.type == array_type)

    @property
    def array_types(self) -> set[ArrayType]:
        """Return unique set of array type (enum) for arrays in dataset."""
        return set(array.type for array in self.values())

    @property
    def array_names(self) -> set[str]:
        """Return unique set of names for arrays in dataset."""
        return set(array.name for array in self.values())

    @property
    def array_quantities(self) -> set[str]:
        """Return unique set of quantities for arrays in dataset."""
        return set(arr.quantity for arr in self.values())

    def current_arrays(self) -> list[DataArray]:
        """Return all Current arrays."""
        return self.arrays_by_type(ArrayType.Current)

    def potential_arrays(self) -> list[DataArray]:
        """Return all Potential arrays."""
        return self.arrays_by_type(ArrayType.Potential)

    def time_arrays(self) -> list[DataArray]:
        """Return all Time arrays."""
        return self.arrays_by_type(ArrayType.Time)

    def freq_arrays(self) -> list[DataArray]:
        """Return all Frequency arrays."""
        return self.arrays_by_type(ArrayType.Frequency)

    def zre_arrays(self) -> list[DataArray]:
        """Return all ZRe arrays."""
        return self.arrays_by_type(ArrayType.ZRe)

    def zim_arrays(self) -> list[DataArray]:
        """Return all ZIm arrays."""
        return self.arrays_by_type(ArrayType.ZIm)

    def aux_input_arrays(self) -> list[DataArray]:
        """Return all AuxInput arrays."""
        return self.arrays_by_type(ArrayType.AuxInput)

    def current_range(self) -> list[str]:
        """Return current range as list of strings."""
        array = self.current_arrays()[-1]

        clr_type = clr.GetClrType(CurrentReading)
        field_info = clr_type.GetField('CurrentRange')

        return [field_info.GetValue(val).ToString() for val in array._psarray]

    def reading_status(self) -> list[str]:
        """Return reading status as list of strings."""
        array = self.current_arrays()[-1]

        clr_type = clr.GetClrType(CurrentReading)
        field_info = clr_type.GetField('ReadingStatus')

        return [field_info.GetValue(val).ToString() for val in array._psarray]

    def timing_status(self) -> list[str]:
        """Return timing status as list of strings."""
        array = self.current_arrays()[-1]

        clr_type = clr.GetClrType(CurrentReading)
        field_info = clr_type.GetField('TimingStatus')

        return [field_info.GetValue(val).ToString() for val in array._psarray]

    def to_dataframe(self) -> pd.DataFrame:
        """Return dataset as pandas dataframe.

        Requires pandas.

        Returns
        -------
        df : pd.DataFrame
            pandas dataframe with all arrays in dataset
        """
        import pandas as pd

        cols, arrays = zip(*[(key, arr.to_list()) for key, arr in self.items() if len(arr)])

        arrays_list = list(arrays)
        arrays_list.append(self.current_range())
        arrays_list.append(self.reading_status())

        cols_list = list(cols)
        cols_list.append('CR')
        cols_list.append('ReadingStatus')

        df = pd.DataFrame(arrays_list, index=cols_list).T

        return df



from __future__ import annotations

from .curve import Curve
from .data_array import DataArray
from .dataset import DataSet
from .eisdata import EISData
from .measurement import DeviceInfo, Measurement
from .peak import Peak

__all__ = [
    'Curve',
    'DataArray',
    'DataSet',
    'DeviceInfo',
    'EISData',
    'Measurement',
    'Peak',
]



from __future__ import annotations

from ._common import Instrument
from .instrument_manager import InstrumentManager, discover
from .instrument_manager_async import InstrumentManagerAsync, discover_async
from .instrument_pool import InstrumentPool
from .instrument_pool_async import InstrumentPoolAsync

__all__ = [
    'discover',
    'discover_async',
    'Instrument',
    'InstrumentManager',
    'InstrumentManagerAsync',
    'InstrumentPool',
    'InstrumentPoolAsync',
]



"""This module contains the public api for classes representing measurement data."""

from __future__ import annotations

from ._data.curve import Curve
from ._data.data_array import ArrayType, DataArray
from ._data.dataset import DataSet
from ._data.eisdata import EISData
from ._data.measurement import DeviceInfo, Measurement
from ._data.peak import Peak

__all__ = [
    'ArrayType',
    'Curve',
    'DataArray',
    'DataSet',
    'DeviceInfo',
    'EISData',
    'Measurement',
    'Peak',
]



from __future__ import annotations

import asyncio
import warnings
from dataclasses import dataclass, field
from math import floor
from typing import Any, Protocol, Sequence

from PalmSens.Comm import enumDeviceType
from System import Action


class Callback(Protocol):
    """Type signature for callback"""

    def __call__(self, new_data: Sequence[dict[str, Any]]): ...


def create_future(clr_task):
    loop = asyncio.get_running_loop()
    future = asyncio.Future()
    callback = Action(lambda: on_completion(future, loop, clr_task))

    clr_task.GetAwaiter().OnCompleted(callback)
    return future


def on_completion(future, loop, task):
    if task.IsFaulted:
        clr_error = task.Exception.GetBaseException()
        future.set_exception(clr_error)
    else:
        loop.call_soon_threadsafe(lambda: future.set_result(task.GetAwaiter().GetResult()))


def firmware_warning(capabilities, /) -> None:
    """Raise warning if firmware is not supported."""
    from pypalmsens import __sdk_version__

    device_type = capabilities.DeviceType
    firmware_version = capabilities.FirmwareVersion
    min_version = capabilities.MinFirmwareVersionRequired

    if not min_version:
        return

    if device_type in (
        enumDeviceType.PalmSens,
        enumDeviceType.EmStat1,
        enumDeviceType.EmStat2,
        enumDeviceType.PalmSens3,
        enumDeviceType.PalmSens4,
        enumDeviceType.EmStat2BP,
        enumDeviceType.EmStat3,
        enumDeviceType.EmStat3P,
        enumDeviceType.EmStat3BP,
    ):
        not_supported = firmware_version < (min_version - 0.01)
    elif device_type in (
        enumDeviceType.EmStatPico,
        enumDeviceType.EmStat4LR,
        enumDeviceType.EmStat4HR,
    ):
        not_supported = int(floor(firmware_version * 10)) < int(floor(min_version * 10))
    else:
        return

    if not_supported:
        warnings.warn(
            (
                f'Device firmware: {firmware_version} on {device_type} '
                f'is not supported by SDK ({__sdk_version__}), '
                f'minimum required firmware version: {min_version}'
            ),
            UserWarning,
            stacklevel=2,
        )


@dataclass
class Instrument:
    """Dataclass holding instrument info."""

    id: str = field(repr=False)
    """Device ID of the instrument."""
    name: str = field(init=False)
    """Name of the instrument."""
    channel: int = field(init=False, default=-1)
    """Channel index if part of a multichannel device.

    Returns -1 if instrument is not part of a multi-channel device."""
    interface: str
    """Type of the connection."""
    device: Any = field(repr=False)
    """Device connection class."""

    def __post_init__(self):
        try:
            idx = self.id.index('CH')
        except ValueError:
            self.name = self.id
        else:
            ch_str = self.id[idx : idx + 5]
            self.channel = int(ch_str[2:])
            self.name = self.id[:idx]

    def __repr__(self):
        args = ''.join(
            (
                f'name={self.name!r}, ',
                f'channel={self.channel}, ' if self.channel > 0 else '',
                f'interface={self.interface!r}',
            )
        )

        return f'{self.__class__.__name__}({args})'



from __future__ import annotations

from enum import Enum

from .data_array import DataArray
from .dataset import DataSet


class EISValueType(Enum):
    X = 0
    """X values"""
    Freq = 1
    """FixedFrequency values"""
    Logf = 2
    """Log(F) values"""
    LogZ = 3
    """Log(Z) values"""
    Edc = 4
    """E DC values"""
    mEdc = 5
    """Mean E DC values"""
    Eac = 6
    """E AC values"""
    Time = 7
    """I values"""
    Idc = 8
    """E values"""
    Iac = 9
    """I AC values"""
    miDC = 10
    """measured I DC values"""
    ZRe = 11
    """Z' values"""
    ZIm = 12
    """-Z'' values"""
    Z = 13
    """Z values"""
    MinPhase = 14
    """-Phase values"""
    Rct = 15
    """RCT values"""
    LogY = 16
    """E or t values"""
    YRe = 17
    """Y' values"""
    YIm = 18
    """Y'' values"""
    Y = 19
    """Y (Admittance) values"""
    Cs = 20
    """Cs (Capacitance in Series) values"""
    CsRe = 21
    """Cs Real values"""
    CsIm = 22
    """-Cs imaginary values"""
    iDCinRange = 23
    AuxInput = 24
    """AuxInput values"""


class EISData:
    """Python wrapper for .NET EISdata class.

    Parameters
    ----------
    pseis
        Reference to .NET EISdata object.
    """

    def __init__(self, *, pseis):
        self._pseis = pseis

    def __repr__(self):
        data = [
            f'title={self.title}',
            f'n_points={self.n_points}',
            f'n_frequencies={self.n_frequencies}',
        ]
        if self.has_subscans:
            data.append(f'n_subscans={self.n_subscans}')

        s = ', '.join(data)
        return f'{self.__class__.__name__}({s})'

    @property
    def title(self) -> str:
        """Tite for EIS data."""
        return self._pseis.Title

    @property
    def frequency_type(self) -> str:
        """Frequency type."""
        return str(self._pseis.FreqType)

    @property
    def scan_type(self) -> str:
        """Scan type."""
        return str(self._pseis.ScanType)

    @property
    def dataset(self) -> DataSet:
        """Dataset which contains multiple arrays of values."""
        return DataSet(psdataset=self._pseis.EISDataSet)

    @property
    def subscans(self) -> list[EISData]:
        """Get list of subscans."""
        return [EISData(pseis=subscan) for subscan in self._pseis.GetSubScans()]

    @property
    def n_points(self) -> int:
        """Number of points (including subscans)."""
        return self._pseis.NPoints

    @property
    def n_frequencies(self) -> int:
        """Number of frequencies."""
        return self._pseis.NFrequencies

    @property
    def n_subscans(self) -> int:
        """Number of subscans."""
        return len(self._pseis.GetSubScans())

    @property
    def x_unit(self) -> str:
        """Unit for array."""
        return self._pseis.XUnit.ToString()

    @property
    def x_quantity(self) -> str:
        """Quantity for array."""
        return self._pseis.XUnit.Quantity

    @property
    def ocp_value(self) -> float:
        """OCP Value."""
        return self._pseis.OCPValue

    @property
    def has_subscans(self) -> bool:
        """Return True if data contains subscans."""
        return self._pseis.HasSubScans

    @property
    def mux_channel(self) -> int:
        """Mux channel."""
        return self._pseis.MuxChannel

    def get_data_for_frequency(self, frequency: int) -> dict[str, DataArray]:
        """Returns dictionary with data per frequency.

        Parameters
        ----------
        frequency : int
            Index of the frequency to retrieve the data for.

        Returns
        -------
        dict[str, DataArray]
            Data are returned as a dictionary keyed by the data type.
        """
        if not (0 <= frequency < self.n_frequencies):
            raise ValueError(f'Frequency must be between 0 and {self.n_frequencies}')

        return {
            str(row.Key): DataArray(psarray=row.Value)
            for row in self._pseis.GetDataArrayVsX(frequency)
        }

    def arrays(self) -> list[DataArray]:
        """Complete list of data arrays."""
        return list(self.dataset.values())

    def current_range(self) -> list[str]:
        """Current ranges for the measurement."""
        return [self._pseis.GetCurrentRange(val).Description for val in range(self.n_points)]

    @property
    def cdc(self) -> str:
        """Gets the CDC circuit for fitting."""
        return self._pseis.CDC

    @property
    def cdc_values(self) -> list[float]:
        """Return values for circuit description code (CDC)."""
        return list(self._pseis.CDCValues)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

import PalmSens.Analysis as PSAnalysis
import System
from PalmSens.Plottables import Curve as PSCurve

from .data_array import DataArray
from .peak import Peak

if TYPE_CHECKING:
    from matplotlib import axes, figure


class Curve:
    """Python wrapper for .NET Curve class.

    Parameters
    ----------
    pscurve
        Reference to .NET curve object.
    """

    def __init__(self, *, pscurve: PSCurve):
        self._pscurve = pscurve

    def __repr__(self):
        return f'{self.__class__.__name__}(title={self.title}, n_points={self.n_points})'

    def copy(self) -> Curve:
        """Return a copy of this curve."""
        return Curve(pscurve=PSCurve(self._pscurve, cloneData=True))

    def smooth(self, smooth_level: int = 0):
        """Smooth the .y_array using a Savitsky-Golay filter with the specified smooth
        level.

        Parameters
        ----------
        smooth_level : int
            The smooth level to be used. -1 = none, 0 = no smooth (spike rejection only),
            1 = 5 points, 2 = 9 points, 3 = 15 points, 4 = 25 points
        """
        success = self._pscurve.Smooth(smoothLevel=smooth_level)
        if not success:
            raise ValueError('Something went wrong.')

    def savitsky_golay(self, window_size: int = 3):
        """Smooth the .y_array using a Savitsky-Golay filter with the specified window
        size.

        (i.e. window size 2 will filter points based on the values of the next/previous 2 points)

        Parameters
        ----------
        window_size : int
            Size of the window
        """
        self._pscurve.SavitskyGolay(windowSize=window_size)

    def find_peaks(
        self,
        min_peak_width: float = 0.1,
        min_peak_height: float = 0.0,
        peak_shoulders: bool = False,
        merge_overlapping_peaks: bool = True,
    ) -> list[Peak]:
        """Find peaks in a curve in all directions.

        CV can have 1 or 2 direction changes.

        Parameters
        ----------
        min_peak_width : float
            Minimum width of the peak in V
        min_peak_height : float
            Minimum height of the peak in uA
        peak_shoulders : bool, optional
            Use alternative peak search algorithm optimized for finding peaks on slopes
        merge_overlapping_peaks : bool, optional
            Two or more peaks that overlap will be identified as a single
            base peak and as shoulder peaks on the base peak.

        Returns
        -------
        peak_list : list[Peak]
        """
        pspeaks = self._pscurve.FindPeaks(
            minPeakWidth=min_peak_width,
            minPeakHeight=min_peak_height,
            peakShoulders=peak_shoulders,
            mergeOverlappingPeaks=merge_overlapping_peaks,
        )

        peaks_list = [Peak(pspeak=peak) for peak in pspeaks]

        return peaks_list

    def find_peaks_semiderivative(
        self,
        min_peak_height: float = 0.0,
    ) -> list[Peak]:
        """
        Find peaks in a curve using the semi-derivative algorithm.

        Used for detecting non-overlapping peaks in LSV and CV curves.
        The peaks are also assigned to the curve, updating `Curve.peaks`.
        Existing peaks are overwritten.

        For more info, see this
        [Wikipedia page](https://en.wikipedia.org/wiki/Neopolarogram).

        Parameters
        ----------
        min_peak_height : float
            Minimum height of the peak in uA

        Returns
        -------
        peak_list : list[Peak]
        """
        dct = System.Collections.Generic.Dictionary[PSCurve, System.Double]()
        dct[self._pscurve] = min_peak_height

        pd = PSAnalysis.SemiDerivativePeakDetection()
        pd.GetNonOverlappingPeaks(dct)

        return self.peaks

    @property
    def max_x(self) -> float:
        """Maximum X value found in this curve."""
        return self._pscurve.MaxX

    @property
    def max_y(self) -> float:
        """Maximum Y value found in this curve."""
        return self._pscurve.MaxY

    @property
    def min_x(self) -> float:
        """Minimum X value found in this curve."""
        return self._pscurve.MinX

    @property
    def min_y(self) -> float:
        """Minimum Y value found in this curve."""
        return self._pscurve.MinY

    @property
    def mux_channel(self) -> int:
        """The corresponding MUX channel number with the curve starting at 0.
        Return -1 when no MUX channel used."""
        return self._pscurve.MuxChannel

    @property
    def n_points(self) -> int:
        """Number of points for this curve."""
        return len(self)

    def __len__(self):
        return self._pscurve.NPoints

    @property
    def ocp_value(self) -> float:
        """OCP value for curve."""
        return self._pscurve.OCPValue

    @property
    def reference_electrode_name(self) -> Union[None, str]:
        """The name of the reference electrode. Return None if not set."""
        if ret := self._pscurve.ReferenceElectrodeName:
            return str(ret)
        return None

    @property
    def reference_electrode_potential(self) -> Union[None, str]:
        """The reference electrode potential offset. Return None if not set."""
        if ret := self._pscurve.ReferenceElectrodePotential:
            return str(ret)
        return None

    @property
    def x_unit(self) -> str:
        """Units for X dimension."""
        return self._pscurve.XUnit.ToString()

    @property
    def x_label(self) -> str:
        """Label for X dimension."""
        return self._pscurve.XUnit.Quantity

    @property
    def y_unit(self) -> str:
        """Units for Y dimension."""
        return self._pscurve.YUnit.ToString()

    @property
    def y_label(self) -> str:
        """Label for Y dimension."""
        return self._pscurve.YUnit.Quantity

    @property
    def z_unit(self) -> Union[None, str]:
        """Units for Z dimension. Returns None if not set."""
        if ret := self._pscurve.ZUnit:
            return ret.ToString()
        return None

    @property
    def z_label(self) -> Union[None, str]:
        """Units for Z dimension. Returns None if not set."""
        if ret := self._pscurve.ZUnit:
            return ret.Quantity
        return None

    @property
    def title(self) -> str:
        """Title for the curve."""
        return self._pscurve.Title

    @title.setter
    def title(self, title: str):
        """Set the title for the curve."""
        self._pscurve.Title = title

    @property
    def peaks(self) -> list[Peak]:
        """Return peaks stored on object."""
        try:
            peaks = [Peak(pspeak=peak) for peak in self._pscurve.Peaks]
        except TypeError:
            peaks = []
        return peaks

    def clear_peaks(self):
        """Clear peaks stored on object."""
        self._pscurve.ClearPeaks()

    @property
    def x_array(self) -> DataArray:
        """Y data for the curve."""
        return DataArray(psarray=self._pscurve.XAxisDataArray)

    @property
    def y_array(self) -> DataArray:
        """Y data for the curve."""
        return DataArray(psarray=self._pscurve.YAxisDataArray)

    def linear_slope(
        self, start: Optional[int] = None, stop: Optional[int] = None
    ) -> tuple[float, float, float]:
        """Calculate linear line parameters for this curve between two indexes.

        current = a + b * x

        Parameters
        ----------
        start : int, optional
            begin index
        stop : int, optional
            end index

        Returns
        -------
        a : float
        b : float
        coefdet : float
            Coefficient of determination (R2)
        """
        if start and stop:
            return self._pscurve.LLS(start, stop)
        else:
            return self._pscurve.LLS()

    def plot(
        self,
        ax: Optional[axes.Axes] = None,
        legend: bool = True,
        **plot_kwargs,
    ) -> figure.Figure:
        """Generate simple plot for this curve using matplotlib.

        Parameters
        ----------
        ax : Optional[axes.Axes]
            Add plot to this ax if specified.
        legend : bool
            If True, add legend.
        plot_kwargs
            These keyword arguments are passed to `ax.plot`.

        Returns
        -------
        fig : fig.Figure
            Matplotlib figure. Use `fig.show()` to render plot.
        """
        import matplotlib.pyplot as plt

        if not ax:
            fig, ax = plt.subplots()

        ax.plot(self.x_array, self.y_array, label=self.title, **plot_kwargs)
        ax.set_xlabel(f'{self.x_label} ({self.x_unit})')
        ax.set_ylabel(f'{self.y_label} ({self.y_unit})')

        if peaks := self.peaks:
            x, y = list(zip(*((peak.x, peak.y) for peak in peaks)))
            ax.scatter(x, y, label='Peaks')

        if legend:
            ax.legend()

        return ax.figure



from __future__ import annotations

from typing import Literal

import attrs
import PalmSens
from PalmSens import Method as PSMethod
from PalmSens import MuxMethod as PSMuxMethod

from pypalmsens._shared import single_to_double

from ._shared import (
    CURRENT_RANGE,
    POTENTIAL_RANGE,
    convert_bools_to_int,
    convert_int_to_bools,
)
from .base import BaseSettings


@attrs.define
class CurrentRange(BaseSettings):
    """Set the autoranging current for a given method."""

    max: CURRENT_RANGE = CURRENT_RANGE.cr_10_mA
    """Maximum current range.

    Use `CURRENT_RANGE` to define the range."""

    min: CURRENT_RANGE = CURRENT_RANGE.cr_1_uA
    """Minimum current range.

    Use `CURRENT_RANGE` to define the range."""

    start: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Start current range.

    Use `CURRENT_RANGE` to define the range."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.Ranging.MaximumCurrentRange = self.max._to_psobj()
        psmethod.Ranging.MinimumCurrentRange = self.min._to_psobj()
        psmethod.Ranging.StartCurrentRange = self.start._to_psobj()

    def _update_params(self, psmethod: PSMethod, /):
        self.max = CURRENT_RANGE._from_psobj(psmethod.Ranging.MaximumCurrentRange)
        self.min = CURRENT_RANGE._from_psobj(psmethod.Ranging.MinimumCurrentRange)
        self.start = CURRENT_RANGE._from_psobj(psmethod.Ranging.StartCurrentRange)


@attrs.define
class PotentialRange(BaseSettings):
    """Set the autoranging potential for a given method."""

    max: POTENTIAL_RANGE = POTENTIAL_RANGE.pr_1_V
    """Maximum potential range.

    Use `POTENTIAL_RANGE` to define the range."""

    min: POTENTIAL_RANGE = POTENTIAL_RANGE.pr_1_mV
    """Minimum potential range.

    Use `POTENTIAL_RANGE` to define the range."""

    start: POTENTIAL_RANGE = POTENTIAL_RANGE.pr_1_V
    """Start potential range.

    Use `POTENTIAL_RANGE` to define the range."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.RangingPotential.MaximumPotentialRange = self.max._to_psobj()
        psmethod.RangingPotential.MinimumPotentialRange = self.min._to_psobj()
        psmethod.RangingPotential.StartPotentialRange = self.start._to_psobj()

    def _update_params(self, psmethod: PSMethod, /):
        self.max = POTENTIAL_RANGE._from_psobj(psmethod.RangingPotential.MaximumPotentialRange)
        self.min = POTENTIAL_RANGE._from_psobj(psmethod.RangingPotential.MinimumPotentialRange)
        self.start = POTENTIAL_RANGE._from_psobj(psmethod.RangingPotential.StartPotentialRange)


@attrs.define
class Pretreatment(BaseSettings):
    """Set the pretreatment settings for a given method."""

    deposition_potential: float = 0.0
    """Deposition potential in V"""

    deposition_time: float = 0.0
    """Deposition time in s"""

    conditioning_potential: float = 0.0
    """Conditioning potential in V"""

    conditioning_time: float = 0.0
    """Conditioning time in s"""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.DepositionPotential = self.deposition_potential
        psmethod.DepositionTime = self.deposition_time
        psmethod.ConditioningPotential = self.conditioning_potential
        psmethod.ConditioningTime = self.conditioning_time

    def _update_params(self, psmethod: PSMethod, /):
        self.deposition_potential = psmethod.DepositionPotential
        self.deposition_time = psmethod.DepositionTime
        self.conditioning_potential = psmethod.ConditioningPotential
        self.conditioning_time = psmethod.ConditioningTime


@attrs.define
class VersusOCP(BaseSettings):
    """Set the versus OCP settings for a given method."""

    mode: int = 0
    """Set versus OCP mode.

    Possible values:
    * 0 = disable versus OCP
    * 1 = vertex 1 potential
    * 2 = vertex 2 potential
    * 3 = vertex 1 & 2 potential
    * 4 = begin potential
    * 5 = begin & vertex 1 potential
    * 6 = begin & vertex 2 potential
    * 7 = begin & vertex 1 & 2 potential
    """

    max_ocp_time: float = 20.0
    """Maximum OCP time in s"""

    stability_criterion: int = 0
    """Stability criterion (potential/time) in mV/s.

    If equal to 0 means no stability criterion.
    If larger than 0, then the value is taken as the stability threshold.
    """

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.OCPmode = self.mode
        psmethod.OCPMaxOCPTime = self.max_ocp_time
        psmethod.OCPStabilityCriterion = self.stability_criterion

    def _update_params(self, psmethod: PSMethod, /):
        self.mode = psmethod.OCPmode
        self.max_ocp_time = psmethod.OCPMaxOCPTime
        self.stability_criterion = psmethod.OCPStabilityCriterion


@attrs.define
class BiPot(BaseSettings):
    """Set the bipot settings for a given method."""

    _mode_t = Literal['constant', 'offset']
    _MODES: tuple[_mode_t, ...] = ('constant', 'offset')

    mode: _mode_t = 'constant'
    """Set the bipotential mode.

    Possible values: `constant` or `offset`"""

    potential: float = 0.0
    """Set the bipotential in V"""

    current_range_max: CURRENT_RANGE = CURRENT_RANGE.cr_10_mA
    """Maximum bipotential current range in mA.

    Use `CURRENT_RANGE` to define the range."""

    current_range_min: CURRENT_RANGE = CURRENT_RANGE.cr_1_uA
    """Minimum bipotential current range.

    Use `CURRENT_RANGE` to define the range."""

    current_range_start: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Start bipotential current range.

    Use `CURRENT_RANGE` to define the range."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        bipot_num = self._MODES.index(self.mode)
        psmethod.BipotModePS = PalmSens.Method.EnumPalmSensBipotMode(bipot_num)
        psmethod.BiPotPotential = self.potential
        psmethod.BipotRanging.MaximumCurrentRange = self.current_range_max._to_psobj()
        psmethod.BipotRanging.MinimumCurrentRange = self.current_range_min._to_psobj()
        psmethod.BipotRanging.StartCurrentRange = self.current_range_start._to_psobj()

    def _update_params(self, psmethod: PSMethod, /):
        self.mode = self._MODES[int(psmethod.BipotModePS)]
        self.potential = psmethod.BiPotPotential
        self.current_range_max = CURRENT_RANGE._from_psobj(
            psmethod.BipotRanging.MaximumCurrentRange
        )
        self.current_range_min = CURRENT_RANGE._from_psobj(
            psmethod.BipotRanging.MinimumCurrentRange
        )
        self.current_range_start = CURRENT_RANGE._from_psobj(
            psmethod.BipotRanging.StartCurrentRange
        )


@attrs.define
class PostMeasurement(BaseSettings):
    """Set the post measurement settings for a given method."""

    cell_on_after_measurement: bool = False
    """Enable/disable cell after measurement."""

    standby_potential: float = 0.0
    """Standby potential (V) for use with cell on after measurement."""

    standby_time: float = 0.0
    """Standby time (s) for use with cell on after measurement."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.CellOnAfterMeasurement = self.cell_on_after_measurement
        psmethod.StandbyPotential = self.standby_potential
        psmethod.StandbyTime = self.standby_time

    def _update_params(self, psmethod: PSMethod, /):
        self.cell_on_after_measurement = psmethod.CellOnAfterMeasurement
        self.standby_potential = psmethod.StandbyPotential
        self.standby_time = psmethod.StandbyTime


@attrs.define
class CurrentLimits(BaseSettings):
    """Set the limit settings for a given method.

    Depending on the method, this will:
    - Abort the measurement
    - Reverse the scan instead (CV)
    - Proceed to the next stage (Mixed Mode)
    """

    max: None | float = None
    """Set limit current max in µA."""

    min: None | float = None
    """Set limit current min in µA."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if self.max is not None:
            psmethod.UseLimitMaxValue = True
            psmethod.LimitMaxValue = self.max
        else:
            psmethod.UseLimitMaxValue = False

        if self.min is not None:
            psmethod.UseLimitMinValue = True
            psmethod.LimitMinValue = self.min
        else:
            psmethod.UseLimitMinValue = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseLimitMaxValue:
            self.max = psmethod.LimitMaxValue
        else:
            self.max = None

        if psmethod.UseLimitMinValue:
            self.min = psmethod.LimitMinValue
        else:
            self.min = None


@attrs.define
class PotentialLimits(BaseSettings):
    """Set the limit settings for a given method.

    Depending on the method, this will:
    - Abort the measurement
    - Proceed to the next stage (Mixed Mode)
    """

    max: None | float = None
    """Set limit potential max in V."""

    min: None | float = None
    """Set limit potential min in V."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if self.max is not None:
            psmethod.UseLimitMaxValue = True
            psmethod.LimitMaxValue = self.max
        else:
            psmethod.UseLimitMaxValue = False

        if self.min is not None:
            psmethod.UseLimitMinValue = True
            psmethod.LimitMinValue = self.min
        else:
            psmethod.UseLimitMinValue = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseLimitMaxValue:
            self.max = psmethod.LimitMaxValue
        else:
            self.max = None

        if psmethod.UseLimitMinValue:
            self.min = psmethod.LimitMinValue
        else:
            self.min = None


@attrs.define
class ChargeLimits(BaseSettings):
    """Set the charge limit settings for a given method."""

    max: None | float = 0.0
    """Set limit charge max in µC."""

    min: None | float = 0.0
    """Set limit charge min in µC."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if self.max is not None:
            psmethod.UseChargeLimitMax = True
            psmethod.ChargeLimitMax = self.max
        else:
            psmethod.UseChargeLimitMax = False

        if self.min is not None:
            psmethod.UseChargeLimitMin = True
            psmethod.ChargeLimitMin = self.min
        else:
            psmethod.UseChargeLimitMin = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseChargeLimitMax:
            self.max = psmethod.ChargeLimitMax
        else:
            self.max = None

        if psmethod.UseChargeLimitMin:
            self.min = psmethod.ChargeLimitMin
        else:
            self.min = None


@attrs.define
class IrDropCompensation(BaseSettings):
    """Set the iR drop compensation settings for a given method."""

    resistance: None | float = None
    """Set the iR compensation resistance in Ω"""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if self.resistance:
            psmethod.UseIRDropComp = True
            psmethod.IRDropCompRes = self.resistance
        else:
            psmethod.UseIRDropComp = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseIRDropComp:
            self.resistance = psmethod.IRDropCompRes
        else:
            self.resistance = None


@attrs.define
class EquilibrationTriggers(BaseSettings):
    """Set the trigger at equilibration settings for a given method.

    If enabled, set one or more digital outputs at the start of
    the equilibration period.
    """

    d0: bool = False
    """If True, enable trigger at d0 high."""

    d1: bool = False
    """If True, enable trigger at d1 high."""

    d2: bool = False
    """If True, enable trigger at d2 high."""

    d3: bool = False
    """If True, enable trigger at d3 high."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if any((self.d0, self.d1, self.d2, self.d3)):
            psmethod.UseTriggerOnEquil = True
            psmethod.TriggerValueOnEquil = convert_bools_to_int(
                (self.d0, self.d1, self.d2, self.d3)
            )
        else:
            psmethod.UseTriggerOnEquil = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseTriggerOnEquil:
            self.d0, self.d1, self.d2, self.d3 = convert_int_to_bools(
                psmethod.TriggerValueOnEquil
            )
        else:
            self.d0 = False
            self.d1 = False
            self.d2 = False
            self.d3 = False


@attrs.define
class MeasurementTriggers(BaseSettings):
    """Set the trigger at measurement settings for a given method.

    If enabled, set one or more digital outputs at the start measurement,
    """

    d0: bool = False
    """If True, enable trigger at d0 high."""

    d1: bool = False
    """If True, enable trigger at d1 high."""

    d2: bool = False
    """If True, enable trigger at d2 high."""

    d3: bool = False
    """If True, enable trigger at d3 high."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        if any((self.d0, self.d1, self.d2, self.d3)):
            psmethod.UseTriggerOnStart = True
            psmethod.TriggerValueOnStart = convert_bools_to_int(
                (self.d0, self.d1, self.d2, self.d3)
            )
        else:
            psmethod.UseTriggerOnEquil = False

    def _update_params(self, psmethod: PSMethod, /):
        if psmethod.UseTriggerOnStart:
            self.d0, self.d1, self.d2, self.d3 = convert_int_to_bools(
                psmethod.TriggerValueOnStart
            )
        else:
            self.d0 = False
            self.d1 = False
            self.d2 = False
            self.d3 = False


@attrs.define
class DelayTriggers(BaseSettings):
    """Set the delayed trigger at measurement settings for a given method.

    If enabled, set one or more digital outputs at the start measurement after a delay,
    """

    delay: float = 0.5
    """Delay in s after the measurement has started.

    The value will be rounded to interval time * number of data points.
    """

    d0: bool = False
    """If True, enable trigger at d0 high."""

    d1: bool = False
    """If True, enable trigger at d1 high."""

    d2: bool = False
    """If True, enable trigger at d2 high."""

    d3: bool = False
    """If True, enable trigger at d3 high."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.TriggerDelayPeriod = self.delay

        if any((self.d0, self.d1, self.d2, self.d3)):
            psmethod.UseTriggerOnDelay = True
            psmethod.TriggerValueOnDelay = convert_bools_to_int(
                (self.d0, self.d1, self.d2, self.d3)
            )
        else:
            psmethod.UseTriggerOnDelay = False

    def _update_params(self, psmethod: PSMethod, /):
        self.delay = psmethod.TriggerDelayPeriod

        if psmethod.UseTriggerOnDelay:
            self.d0, self.d1, self.d2, self.d3 = convert_int_to_bools(
                psmethod.TriggerValueOnDelay
            )
        else:
            self.d0 = False
            self.d1 = False
            self.d2 = False
            self.d3 = False


@attrs.define
class Multiplexer(BaseSettings):
    """Set the multiplexer settings for a given method."""

    _mode_t = Literal['none', 'consecutive', 'alternate']
    _MODES: tuple[_mode_t, ...] = ('none', 'consecutive', 'alternate')

    mode: _mode_t = 'none'
    """Set multiplexer mode.

    Possible values:
    * 'none' = No multiplexer (disable)
    * 'consecutive
    * 'alternate
    """

    channels: list[int] = attrs.field(factory=list)
    """Set multiplexer channels

    This is defined as a list of indexes for which channels to enable (max 128).
    For example, [0,3,7]. In consecutive mode all selections are valid.

    In alternating mode the first channel must be selected and all other
    channels should be consecutive i.e. (channel 1, channel 2, channel 3 and so on).
    """
    connect_sense_to_working_electrode: bool = False
    """Connect the sense electrode to the working electrode. Default is False."""

    combine_reference_and_counter_electrodes: bool = False
    """Combine the reference and counter electrodes. Default is False."""

    use_channel_1_reference_and_counter_electrodes: bool = False
    """Use channel 1 reference and counter electrodes for all working electrodes. Default is False."""

    set_unselected_channel_working_electrode: int = 0
    """Set the unselected channel working electrode to 0 = Disconnected / floating, 1 = Ground, 2 = Standby potential. Default is 0."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        # Create a mux8r2 multiplexer settings settings object
        mux_mode = self._MODES.index(self.mode) - 1
        psmethod.MuxMethod = PSMuxMethod(mux_mode)

        # disable all mux channels (range 0-127)
        for i in range(len(psmethod.UseMuxChannel)):
            psmethod.UseMuxChannel[i] = False

        # set the selected mux channels
        for i in self.channels:
            psmethod.UseMuxChannel[i - 1] = True

        psmethod.MuxSett.ConnSEWE = self.connect_sense_to_working_electrode
        psmethod.MuxSett.ConnectCERE = self.combine_reference_and_counter_electrodes
        psmethod.MuxSett.CommonCERE = self.use_channel_1_reference_and_counter_electrodes
        psmethod.MuxSett.UnselWE = PSMethod.MuxSettings.UnselWESetting(
            self.set_unselected_channel_working_electrode
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.mode = self._MODES[int(psmethod.MuxMethod) + 1]

        self.channels = [
            i + 1 for i in range(len(psmethod.UseMuxChannel)) if psmethod.UseMuxChannel[i]
        ]

        self.connect_sense_to_working_electrode = psmethod.MuxSett.ConnSEWE
        self.combine_reference_and_counter_electrodes = psmethod.MuxSett.ConnectCERE
        self.use_channel_1_reference_and_counter_electrodes = psmethod.MuxSett.CommonCERE
        self.set_unselected_channel_working_electrode = int(psmethod.MuxSett.UnselWE)


@attrs.define
class DataProcessing(BaseSettings):
    """Set the data processing settings for a given method."""

    smooth_level: int = 0
    """Set the default curve post processing filter.

    Possible values:
    * -1 = no filter
    *  0 = spike rejection
    *  1 = spike rejection + Savitsky-golay window 5
    *  2 = spike rejection + Savitsky-golay window 9
    *  3 = spike rejection + Savitsky-golay window 15
    *  4 = spike rejection + Savitsky-golay window 25
    """

    min_height: float = 0.0
    """Determines the minimum peak height in µA for peak finding.

    Peaks lower than this value are neglected."""
    min_width: float = 0.1
    """The minimum peak width for peak finding.

    The value is in the unit of the curves X axis (V).
    Peaks narrower than this value are neglected (default: 0.1 V)."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.SmoothLevel = self.smooth_level
        psmethod.MinPeakHeight = self.min_height
        psmethod.MinPeakWidth = self.min_width

    def _update_params(self, psmethod: PSMethod, /):
        self.smooth_level = psmethod.SmoothLevel
        self.min_width = single_to_double(psmethod.MinPeakWidth)
        self.min_height = single_to_double(psmethod.MinPeakHeight)


@attrs.define
class General(BaseSettings):
    """Sets general/other settings for a given method."""

    save_on_internal_storage: bool = False
    """Save on internal storage."""

    use_hardware_sync: bool = False
    """Use hardware synchronization with other channels/instruments."""

    notes: str = ''
    """Add some user notes for use with this technique."""

    power_frequency: Literal[50, 60] = 50
    """Set the DC mains filter in Hz.

    Adjusts sampling on instrument to account for mains frequency.
    Set to 50 Hz or 60 Hz depending on your region (default: 50)."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        psmethod.SaveOnDevice = self.save_on_internal_storage
        psmethod.UseHWSync = self.use_hardware_sync
        psmethod.Notes = self.notes
        psmethod.PowerFreq = self.power_frequency

    def _update_params(self, psmethod: PSMethod, /):
        self.save_on_internal_storage = psmethod.SaveOnDevice
        self.use_hardware_sync = psmethod.UseHWSync
        self.notes = psmethod.Notes
        self.power_frequency = psmethod.PowerFreq



from __future__ import annotations

import atexit
import platform
from importlib.resources import files
from pathlib import Path

from pythonnet import load, unload

PSSDK_DIR = files('pypalmsens._pssdk.mono')

# runtime must be imported before clr is loaded
load('coreclr', runtime_config=str(PSSDK_DIR / 'runtimeconfig.json'))

import clr  # noqa: E402

core_dll = PSSDK_DIR / 'PalmSens.Core.dll'
core_linux_dll = PSSDK_DIR / 'PalmSens.Core.Linux.dll'

RUNTIME_DIR = files('pypalmsens._runtimes')

system = platform.system()  # Windows, Linux, Darwin
machine = platform.machine()  # AMD64, x86_64, arm64

# Select the correct version of the SerialPort library
# To use serial devices the correct version of the libSystem.IO.Ports.Native.so
# library must be loaded to into pythonnet.
PLATFORMS = {
    ('Linux', 'x86_64'): 'linux-x64',
    ('Linux', 'arm'): 'linux-arm',
    ('Linux', 'aarch'): 'linux-arm',
    ('Linux', 'arm64'): 'linux-arm64',
    ('Linux', 'aarch64'): 'linux-arm64',
    ('Darwin', 'arm64'): 'osx-arm64',
    ('Darwin', 'x86_64'): 'osx-x64',
}

plat = PLATFORMS[system, machine]

ioports_dll = RUNTIME_DIR / plat / 'native' / 'System.IO.Ports.dll'

assert isinstance(core_dll, Path)
assert isinstance(core_linux_dll, Path)
assert isinstance(ioports_dll, Path)

# This dll contains the classes in which the data is stored
clr.AddReference(str(core_dll.with_suffix('')))

# This dll is used to load your session file
clr.AddReference(str(core_linux_dll.with_suffix('')))

clr.AddReference(str(ioports_dll.with_suffix('')))

clr.AddReference('System')

from PalmSens.Core.Linux import CoreDependencies  # noqa: E402
from System import Diagnostics  # noqa: E402

CoreDependencies.Init()

sdk_version = Diagnostics.FileVersionInfo.GetVersionInfo(str(core_dll)).ProductVersion

atexit.register(unload)



from __future__ import annotations

from typing import Literal

import attrs
import PalmSens.Techniques as PSTechniques
from PalmSens import FixedCurrentRange as PSFixedCurrentRange
from PalmSens import Method as PSMethod
from PalmSens.Techniques.Impedance import enumFrequencyType, enumScanType

from . import mixins
from ._shared import (
    CURRENT_RANGE,
    ELevel,
    ILevel,
    get_extra_value_mask,
    set_extra_value_mask,
)
from .base import BaseTechnique


@attrs.define
class CyclicVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create cyclic voltammetry method parameters."""

    _id = 'cv'

    equilibration_time: float = 0.0
    """Equilibration time in s"""

    begin_potential: float = -0.5
    """Begin potential in V"""

    vertex1_potential: float = 0.5
    """Vertex 1 potential in V"""

    vertex2_potential: float = -0.5
    """Vertex 2 potential in V"""

    step_potential: float = 0.1
    """Step potential in V"""

    scanrate: float = 1.0
    """Scan rate in V/s"""

    n_scans: int = 1
    """Number of scans"""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with fast cyclic voltammetry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.Vtx1Potential = self.vertex1_potential
        psmethod.Vtx2Potential = self.vertex2_potential
        psmethod.StepPotential = self.step_potential
        psmethod.Scanrate = self.scanrate
        psmethod.nScans = self.n_scans

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.vertex1_potential = psmethod.Vtx1Potential
        self.vertex2_potential = psmethod.Vtx2Potential
        self.step_potential = psmethod.StepPotential
        self.scanrate = psmethod.Scanrate
        self.n_scans = psmethod.nScans

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class FastCyclicVoltammetry(
    BaseTechnique,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.PostMeasurementMixin,
    mixins.IrDropCompensationMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create fast cyclic voltammetry method parameters."""

    _id = 'fcv'

    current_range: CURRENT_RANGE = CURRENT_RANGE.cr_1_uA
    """Fixed current range."""

    equilibration_time: float = 0.0
    """Equilibration time in s"""

    begin_potential: float = -0.5
    """Begin potential in V"""

    vertex1_potential: float = 0.5
    """Vertex 1 potential in V"""

    vertex2_potential: float = -0.5
    """Vertex 2 potential in V"""

    step_potential: float = 0.01
    """Step potential in V"""

    scanrate: float = 500.0
    """Scan rate in V/s"""

    n_scans: int = 1
    """Number of scans"""

    n_avg_scans: int = 1
    """Number of scans to be averaged."""

    n_equil_scans: int = 1
    """Number of equilibration scans."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with fast cyclic voltammetry settings."""

        psmethod.Ranging = PSFixedCurrentRange(self.current_range._to_psobj())
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.Vtx1Potential = self.vertex1_potential
        psmethod.Vtx2Potential = self.vertex2_potential
        psmethod.StepPotential = self.step_potential
        psmethod.Scanrate = self.scanrate
        psmethod.nScans = self.n_scans
        psmethod.nAvgScans = self.n_avg_scans
        psmethod.nEqScans = self.n_equil_scans

    def _update_params(self, psmethod: PSMethod, /):
        self.current_range = CURRENT_RANGE._from_psobj(psmethod.Ranging.StartCurrentRange)
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.vertex1_potential = psmethod.Vtx1Potential
        self.vertex2_potential = psmethod.Vtx2Potential
        self.step_potential = psmethod.StepPotential
        self.scanrate = psmethod.Scanrate
        self.n_scans = psmethod.nScans
        self.n_avg_scans = psmethod.nAvgScans
        self.n_equil_scans = psmethod.nEqScans


@attrs.define
class ACVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.PostMeasurementMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create AC Voltammetry method parameters."""

    _id = 'acv'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    ac_potential: float = 0.01
    """Sine wave amplitude in V as rms value."""

    frequency: float = 100.0
    """AC frequency in HZ."""

    scanrate: float = 1.0
    """Scan rate in V/s."""

    measure_dc_current: bool = False
    """Measure the DC current seperately."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with linear sweep settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.EndPotential = self.end_potential
        psmethod.StepPotential = self.step_potential
        psmethod.Frequency = self.frequency
        psmethod.SineWaveAmplitude = self.ac_potential
        psmethod.MeasureDCcurrent = self.measure_dc_current
        psmethod.Scanrate = self.scanrate

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.end_potential = psmethod.EndPotential
        self.step_potential = psmethod.StepPotential
        self.ac_potential = psmethod.SineWaveAmplitude
        self.frequency = psmethod.Frequency
        self.scanrate = psmethod.Scanrate
        self.measure_dc_current = psmethod.MeasureDCcurrent


@attrs.define
class LinearSweepVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create linear sweep method parameters."""

    _id = 'lsv'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    scanrate: float = 1.0
    """Scan rate in V/s."""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with linear sweep settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.EndPotential = self.end_potential
        psmethod.StepPotential = self.step_potential
        psmethod.Scanrate = self.scanrate

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.end_potential = psmethod.EndPotential
        self.step_potential = psmethod.StepPotential
        self.scanrate = psmethod.Scanrate

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class SquareWaveVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create square wave method parameters."""

    _id = 'swv'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    frequency: float = 10.0
    """Frequency in Hz."""

    amplitude: float = 0.05
    """Amplitude in V as half peak-to-peak value."""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    record_forward_and_reverse_currents: bool = False
    """Record forward and reverse currents"""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with square wave voltammetry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.EndPotential = self.end_potential
        psmethod.StepPotential = self.step_potential
        psmethod.Frequency = self.frequency
        psmethod.PulseAmplitude = self.amplitude

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
            record_forward_and_reverse_currents=self.record_forward_and_reverse_currents,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.end_potential = psmethod.EndPotential
        self.step_potential = psmethod.StepPotential
        self.frequency = psmethod.Frequency
        self.amplitude = psmethod.PulseAmplitude

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
            'record_forward_and_reverse_currents',
        ):
            setattr(self, key, msk[key])


@attrs.define
class DifferentialPulseVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create differential pulse voltammetry method parameters."""

    _id = 'dpv'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    pulse_potential: float = 0.05
    """Pulse potential in V."""

    pulse_time: float = 0.01
    """Pulse time in s."""

    scan_rate: float = 1.0
    """Scan rate (potential/time) in V/s."""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with linear sweep settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.EndPotential = self.end_potential
        psmethod.StepPotential = self.step_potential
        psmethod.PulsePotential = self.pulse_potential
        psmethod.PulseTime = self.pulse_time
        psmethod.Scanrate = self.scan_rate

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.end_potential = psmethod.EndPotential
        self.step_potential = psmethod.StepPotential
        self.pulse_potential = psmethod.PulsePotential
        self.pulse_time = psmethod.PulseTime
        self.scan_rate = psmethod.Scanrate

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class NormalPulseVoltammetry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create normal pulse voltammetry method parameters."""

    _id = 'npv'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    pulse_time: float = 0.01
    """Pulse time in s."""

    scan_rate: float = 1.0
    """Scan rate (potential/time) in V/s."""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with normal pulse voltammetry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.BeginPotential = self.begin_potential
        psmethod.EndPotential = self.end_potential
        psmethod.StepPotential = self.step_potential
        psmethod.PulseTime = self.pulse_time
        psmethod.Scanrate = self.scan_rate

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.begin_potential = psmethod.BeginPotential
        self.end_potential = psmethod.EndPotential
        self.step_potential = psmethod.StepPotential
        self.pulse_time = psmethod.PulseTime
        self.scan_rate = psmethod.Scanrate

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class ChronoAmperometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.ChargeLimitsMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create chrono amperometry method parameters."""

    _id = 'ad'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    interval_time: float = 0.1
    """Interval time in s."""

    potential: float = 0.0
    """Potential in V."""

    run_time: float = 1.0
    """Run time in s."""

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with chrono amperometry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.IntervalTime = self.interval_time
        psmethod.Potential = self.potential
        psmethod.RunTime = self.run_time

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.interval_time = psmethod.IntervalTime
        self.potential = psmethod.Potential
        self.run_time = psmethod.RunTime

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class FastAmperometry(
    BaseTechnique,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.ChargeLimitsMixin,
    mixins.IrDropCompensationMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create fast amperometry method parameters."""

    _id = 'fam'

    current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_nA
    """Fixed current range."""

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    equilibration_potential: float = 1.0
    """Equilibration potential in V."""

    interval_time: float = 0.1
    """Interval time in s."""

    potential: float = 0.5
    """Potential in V."""

    run_time: float = 1.0
    """Run time in s."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with fast amperometry settings."""
        psmethod.Ranging = PSFixedCurrentRange(self.current_range._to_psobj())
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.EqPotentialFA = self.equilibration_potential
        psmethod.IntervalTime = self.interval_time
        psmethod.Potential = self.potential
        psmethod.RunTime = self.run_time

    def _update_params(self, psmethod: PSMethod, /):
        self.current_range = CURRENT_RANGE._from_psobj(psmethod.Ranging.StartCurrentRange)
        self.equilibration_time = psmethod.EquilibrationTime
        self.equilibration_potential = psmethod.EqPotentialFA
        self.interval_time = psmethod.IntervalTime
        self.potential = psmethod.Potential
        self.run_time = psmethod.RunTime


@attrs.define
class MultiStepAmperometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.IrDropCompensationMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create multi-step amperometry method parameters."""

    _id = 'ma'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    interval_time: float = 0.1
    """Interval time in s."""

    n_cycles: float = 1
    """Number of cycles."""

    levels: list[ELevel] = attrs.field(factory=lambda: [ELevel()])
    """List of levels.

    Use `ELevel()` to create levels.
    """

    enable_bipot_current: bool = False
    """Enable bipot current."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with multistep amperometry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.IntervalTime = self.interval_time
        psmethod.nCycles = self.n_cycles
        psmethod.Levels.Clear()

        if not self.levels:
            raise ValueError('At least one level must be specified.')

        for level in self.levels:
            psmethod.Levels.Add(level.to_psobj())

        psmethod.UseSelectiveRecord = any(level.record for level in self.levels)
        psmethod.UseLimits = any(level.use_limits for level in self.levels)

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
            enable_bipot_current=self.enable_bipot_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.interval_time = psmethod.IntervalTime
        self.n_cycles = psmethod.nCycles

        self.levels = [ELevel.from_psobj(pslevel) for pslevel in psmethod.Levels]

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
            'enable_bipot_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class PulsedAmperometricDetection(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.BiPotMixin,
    mixins.PostMeasurementMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DelayTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create pulsed amperometric detection method parameters."""

    _id = 'pad'
    _mode_t = Literal['dc', 'pulse', 'differential']
    _MODES: tuple[_mode_t, ...] = ('dc', 'pulse', 'differential')

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    potential: float = 0.5
    """Potential in V."""

    pulse_potential: float = 0.05
    """Pulse potential in V."""

    pulse_time: float = 0.01
    """Pulse time in s."""

    mode: _mode_t = 'dc'
    """Measurement mode.

    - dc: Measurement is performed at potential (E dc)
    - pulse: measurement is performed at pulse potential (E pulse)
    - differential: measurement is (pulse - dc)
    """

    interval_time: float = 0.1
    """Interval time in s."""

    run_time: float = 10.0
    """Run time in s."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with pulsed amperometric detection settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.IntervalTime = self.interval_time
        psmethod.PulseTime = self.pulse_time
        psmethod.PulsePotentialAD = self.pulse_potential
        psmethod.Potential = self.potential
        psmethod.RunTime = self.run_time

        mode = self._MODES.index(self.mode) + 1
        psmethod.tMode = PSTechniques.PulsedAmpDetection.enumMode(mode)

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.interval_time = psmethod.IntervalTime
        self.potential = psmethod.Potential
        self.pulse_potential = psmethod.PulsePotentialAD
        self.pulse_time = psmethod.PulseTime
        self.run_time = psmethod.RunTime

        self.mode = self._MODES[int(psmethod.tMode) - 1]


@attrs.define
class OpenCircuitPotentiometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.PotentialLimitsMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create open circuit potentiometry method parameters."""

    _id = 'ocp'

    interval_time: float = 0.1
    """Interval time in s."""

    run_time: float = 1.0
    """Run time in s."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_we_current: bool = False
    """Record working electrode current."""

    record_we_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_1_uA
    """Record working electrode current range.

    Use `CURRENT_RANGE` to define the range."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with open circuit potentiometry settings."""
        psmethod.IntervalTime = self.interval_time
        psmethod.RunTime = self.run_time
        psmethod.AppliedCurrentRange = self.record_we_current_range._to_psobj()

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_we_current=self.record_we_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.interval_time = psmethod.IntervalTime
        self.run_time = psmethod.RunTime
        self.record_we_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_we_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class ChronoPotentiometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.PotentialLimitsMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create potentiometry method parameters."""

    _id = 'pot'

    current: float = 0.0
    """The current to apply in the given current range.

    Note that this value acts as a multiplier in the applied current range.

    So if 10 uA is the applied current range and 1.5 is given as current value,
    the applied current will be 15 uA."""

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    interval_time: float = 0.1
    """Interval time in s (default: 0.1)"""

    run_time: float = 1.0
    """Run time in s."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_current: bool = False
    """Record working electrode current."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with chronopotentiometry settings."""
        psmethod.Current = self.current
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psmethod.IntervalTime = self.interval_time
        psmethod.RunTime = self.run_time

        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_current=self.record_we_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.current = psmethod.Current
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)
        self.interval_time = psmethod.IntervalTime
        self.run_time = psmethod.RunTime

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class StrippingChronoPotentiometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create stripping potentiometry method parameters.

    If the stripping current is set to 0, then chemical stripping is performed,
    otherwise it is chemical constant current stripping.
    The applicable range is +- 0.001 microampere to +- 2 milliampere.
    """

    _id = 'scp'

    current: float = 0.0
    """The stripping current to apply in the given current range.

    Note that this value acts as a multiplier in the applied current range.

    So if 10 uA is the applied current range and 1.5 is given as current value,
    the applied current will be 15 uA."""

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    end_potential: float = 0.0
    """Potential in V where measurement ends."""

    measurement_time: float = 1.0
    """Measurement time in s (default: 1.0)"""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with stripping chrono potentiometry settings."""
        psmethod.Current = self.current
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psmethod.MeasurementTime = self.measurement_time
        psmethod.EndPotential = self.end_potential

        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()

    def _update_params(self, psmethod: PSMethod, /):
        self.current = psmethod.Current
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)
        self.measurement_time = psmethod.MeasurementTime
        self.end_potential = psmethod.EndPotential


@attrs.define
class LinearSweepPotentiometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.PotentialLimitsMixin,
    mixins.MeasurementTriggersMixin,
    mixins.DelayTriggersMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create linear sweep potentiometry method parameters."""

    _id = 'lsp'

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    current_begin: float = -1.0
    """Current applied at beginning of measurement.

    This value is multiplied by the defined current range."""

    current_end: float = 1.0
    """Current applied at end of measurement.

    This value is multiplied by the defined current range."""

    current_step: float = 0.01
    """Current step.

    This value is multiplied by the defined current range."""

    scan_rate: float = 1.0
    """The applied scan rate.

    This value is multiplied by the defined current range."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_we_current: bool = False
    """Record working electrode current."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with lineas sweep potentiometry settings."""
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()

        psmethod.BeginCurrent = self.current_begin
        psmethod.EndCurrent = self.current_end
        psmethod.StepCurrent = self.current_step
        psmethod.ScanrateG = self.scan_rate

        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_we_current=self.record_we_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)

        self.current_begin = psmethod.BeginCurrent
        self.current_end = psmethod.EndCurrent
        self.current_step = psmethod.StepCurrent
        self.scan_rate = psmethod.ScanrateG

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_we_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class MultiStepPotentiometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.PotentialLimitsMixin,
    mixins.DataProcessingMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create multi-step potentiometry method parameters."""

    _id = 'mp'

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_1_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    interval_time: float = 0.1
    """Interval time in s."""

    n_cycles: float = 1
    """Number of cycles."""

    levels: list[ILevel] = attrs.field(factory=lambda: [ILevel()])
    """List of levels.

    Use `ILevel()` to create levels.
    """

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_we_current: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with multistep potentiometry settings."""
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psmethod.IntervalTime = self.interval_time
        psmethod.nCycles = self.n_cycles
        psmethod.Levels.Clear()

        if not self.levels:
            raise ValueError('At least one level must be specified.')

        for level in self.levels:
            psmethod.Levels.Add(level.to_psobj())

        psmethod.UseSelectiveRecord = any(level.record for level in self.levels)
        psmethod.UseLimits = any(level.use_limits for level in self.levels)

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_we_current=self.record_we_current,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)

        self.interval_time = psmethod.IntervalTime
        self.n_cycles = psmethod.nCycles

        self.levels = [ILevel.from_psobj(pslevel) for pslevel in psmethod.Levels]

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_we_current',
        ):
            setattr(self, key, msk[key])


@attrs.define
class ChronoCoulometry(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.CurrentLimitsMixin,
    mixins.ChargeLimitsMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create linear sweep method parameters."""

    _id = 'cc'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    interval_time: float = 0.1
    """Interval time in s."""

    step1_potential: float = 0.5
    """Potential applied during first step in V."""

    step1_run_time: float = 5.0
    """Run time for the first step."""

    step2_potential: float = 0.5
    """Potential applied during second step in V."""

    step2_run_time: float = 5.0
    """Run time for the second step."""

    bandwidth: None | float = None
    """Override bandwidth on MethodSCRIPT devices if set."""

    record_auxiliary_input: bool = False
    """Record auxiliary input."""

    record_cell_potential: bool = False
    """Record cell potential.

    Counter electrode vs ground."""

    record_we_potential: bool = False
    """Record applied working electrode potential.

    Reference electrode vs ground."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with chrono coulometry settings."""
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.IntervalTime = self.interval_time

        psmethod.EFirstStep = self.step1_potential
        psmethod.ESecondStep = self.step2_potential
        psmethod.TFirstStep = self.step1_run_time
        psmethod.TSecondStep = self.step2_run_time

        if self.bandwidth is not None:
            psmethod.OverrideBandwidth = True
            psmethod.Bandwidth = self.bandwidth

        set_extra_value_mask(
            obj=psmethod,
            record_auxiliary_input=self.record_auxiliary_input,
            record_cell_potential=self.record_cell_potential,
            record_we_potential=self.record_we_potential,
        )

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.interval_time = psmethod.IntervalTime
        self.step1_potential = psmethod.EFirstStep
        self.step2_potential = psmethod.ESecondStep
        self.step1_run_time = psmethod.TFirstStep
        self.step2_run_time = psmethod.TSecondStep

        if psmethod.OverrideBandwidth:
            self.bandwidth = psmethod.Bandwidth

        msk = get_extra_value_mask(psmethod)

        for key in (
            'record_auxiliary_input',
            'record_cell_potential',
            'record_we_potential',
        ):
            setattr(self, key, msk[key])


@attrs.define
class ElectrochemicalImpedanceSpectroscopy(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.PostMeasurementMixin,
    mixins.MeasurementTriggersMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create potentiometry method parameters."""

    _id = 'eis'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    dc_potential: float = 0.0
    """DC potential in V."""

    ac_potential: float = 0.01
    """AC potential in V RMS."""

    n_frequencies: int = 11
    """Number of frequencies."""

    max_frequency: float = 1e5
    """Maximum frequency in Hz."""

    min_frequency: float = 1e3
    """Minimum frequency in Hz."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with electrochemical impedance spectroscopy settings."""
        psmethod.ScanType = enumScanType.Fixed
        psmethod.FreqType = enumFrequencyType.Scan
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.Potential = self.dc_potential
        psmethod.Eac = self.ac_potential
        psmethod.nFrequencies = self.n_frequencies
        psmethod.MaxFrequency = self.max_frequency
        psmethod.MinFrequency = self.min_frequency

    def _update_params(self, psmethod: PSMethod, /):
        self.equilibration_time = psmethod.EquilibrationTime
        self.dc_potential = psmethod.Potential
        self.ac_potential = psmethod.Eac
        self.n_frequencies = psmethod.nFrequencies
        self.max_frequency = psmethod.MaxFrequency
        self.min_frequency = psmethod.MinFrequency


@attrs.define
class FastImpedanceSpectroscopy(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.VersusOCPMixin,
    mixins.PostMeasurementMixin,
    mixins.MeasurementTriggersMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.GeneralMixin,
):
    """Create fast impedance spectroscopy method parameters."""

    _id = 'fis'

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    interval_time: float = 0.1
    """Interval time in s."""

    run_time: float = 10.0
    """Run time in s."""

    dc_potential: float = 0.0
    """Potential applied during measurement in V."""

    ac_potential: float = 0.01
    """Potential amplitude in V (rms)."""

    frequency: float = 50000.0
    """Frequency in Hz."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with fas impedance spectroscopy settings."""
        psmethod.Eac = self.ac_potential
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.FixedFrequency = self.frequency
        psmethod.IntervalTime = self.interval_time
        psmethod.Potential = self.dc_potential
        psmethod.RunTime = self.run_time

    def _update_params(self, psmethod: PSMethod, /):
        self.ac_potential = psmethod.Eac
        self.equilibration_time = psmethod.EquilibrationTime
        self.frequency = psmethod.FixedFrequency
        self.interval_time = psmethod.IntervalTime
        self.dc_potential = psmethod.Potential
        self.run_time = psmethod.RunTime


@attrs.define
class GalvanostaticImpedanceSpectroscopy(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.EquilibrationTriggersMixin,
    mixins.MeasurementTriggersMixin,
    mixins.MultiplexerMixin,
    mixins.GeneralMixin,
):
    """Create potentiometry method parameters."""

    _id = 'gis'

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    equilibration_time: float = 0.0
    """Equilibration time in s."""

    ac_current: float = 0.01
    """AC current in applied current range RMS."""

    dc_current: float = 0.0
    """DC current in applied current range."""

    n_frequencies: int = 11
    """Number of frequencies."""

    max_frequency: float = 1e5
    """Maximum frequency in Hz."""

    min_frequency: float = 1e3
    """Minimum frequency in Hz."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with galvanic impedance spectroscopy settings."""

        psmethod.ScanType = enumScanType.Fixed
        psmethod.FreqType = enumFrequencyType.Scan
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psmethod.EquilibrationTime = self.equilibration_time
        psmethod.Iac = self.ac_current
        psmethod.Idc = self.dc_current
        psmethod.nFrequencies = self.n_frequencies
        psmethod.MaxFrequency = self.max_frequency
        psmethod.MinFrequency = self.min_frequency

    def _update_params(self, psmethod: PSMethod, /):
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)
        self.equilibration_time = psmethod.EquilibrationTime
        self.ac_current = psmethod.Iac
        self.dc_current = psmethod.Idc
        self.n_frequencies = psmethod.nFrequencies
        self.max_frequency = psmethod.MaxFrequency
        self.min_frequency = psmethod.MinFrequency


@attrs.define
class FastGalvanostaticImpedanceSpectroscopy(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PotentialRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.GeneralMixin,
):
    """Create fast galvanostatic impededance spectroscopy method parameters."""

    _id = 'fgis'

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    run_time: float = 10.0
    """Run time in s."""

    interval_time: float = 0.1
    """Interval time in s."""

    ac_current: float = 0.01
    """AC current in applied current range RMS.

    This value is multiplied by the applied current range."""

    dc_current: float = 0.0
    """DC current in applied current range.

    This value is multiplied by the applied current range."""

    frequency: float = 50000.0
    """Frequency in Hz."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with fast galvanic impedance spectroscopy settings."""
        psmethod.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psmethod.Iac = self.ac_current
        psmethod.Idc = self.dc_current
        psmethod.FixedFrequency = self.frequency
        psmethod.RunTime = self.run_time
        psmethod.IntervalTime = self.interval_time

    def _update_params(self, psmethod: PSMethod, /):
        self.applied_current_range = CURRENT_RANGE._from_psobj(psmethod.AppliedCurrentRange)
        self.ac_current = psmethod.Iac
        self.dc_current = psmethod.Idc
        self.frequency = psmethod.FixedFrequency
        self.run_time = psmethod.RunTime
        self.interval_time = psmethod.IntervalTime


@attrs.define
class MethodScript(BaseTechnique):
    """Create a method script sandbox object."""

    _id = 'ms'

    script: str = """e
wait 100m
if 1 < 2
    send_string "Hello world"
endif

"""
    """Script to run.

    For more info on MethodSCRIPT, see:
        https://www.palmsens.com/methodscript/ for more information."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with MethodScript."""
        psmethod.MethodScript = self.script

    def _update_params(self, psmethod: PSMethod, /):
        self.script = psmethod.MethodScript



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from PalmSens.Analysis import Peak as PSPeak

    from .curve import Curve


class Peak:
    """Python wrapper for .NET Peak class.

    Parameters
    ----------
    pspeak : PalmSens.Analysis.Peak
        Reference to .NET Peak object.
    """

    def __init__(self, *, pspeak: PSPeak):
        self._pspeak = pspeak

        self._curve: Curve | None = None

    def __repr__(self):
        x_unit = self.x_unit
        y_unit = self.y_unit

        return (
            f'{self.__class__.__name__}('
            f'x={self.x:g} {x_unit}, '
            f'y={self.y:g} {y_unit}, '
            f'y_offset={self.y_offset:g} {y_unit}, '
            f'area={self.area:g} {x_unit}{y_unit}, '
            f'width={self.width:g} {x_unit})'
        )

    @property
    def curve(self) -> Curve:
        """Parent curve associated with Peak."""
        from .curve import Curve

        if not self._curve:
            self._curve = Curve(pscurve=self._pspeak.Curve)
        return self._curve

    @property
    def curve_title(self) -> str:
        """Title of parent curve."""
        return self.curve.title

    @property
    def x_unit(self) -> str:
        """Units of X axis"""
        return self.curve.x_unit

    @property
    def y_unit(self) -> str:
        """Units for Y axis"""
        return self.curve.y_unit

    @property
    def analyte_name(self) -> str:
        """Name of analyte."""
        return self._pspeak.AnalyteName

    @analyte_name.setter
    def analyte_name(self, name: str):
        """Set name of analyte."""
        self._pspeak.set_AnalyteName(name)

    @property
    def area(self) -> float:
        """Area of the peak."""
        return self._pspeak.Area

    @property
    def label(self) -> str:
        """Formatted label for the peak value."""
        return self._pspeak.Label

    @property
    def left_index(self) -> int:
        """Left side of the peaks baseline as index number of the curve."""
        return self._pspeak.LeftIndex

    @property
    def left_x(self) -> float:
        """X of the left side of the peak baseline."""
        return self._pspeak.LeftX

    @property
    def left_y(self) -> float:
        """Y of the left side of the peak baseline."""
        return self._pspeak.LeftY

    @property
    def maximum_of_derivative_neg(self) -> float:
        """Maximum derivative of the negative slope of the peak."""
        return self._pspeak.MaximumOfDerivativeNeg

    @property
    def maximum_of_derivative_pos(self) -> float:
        """Maximum derivative of the positive slope of the peak."""
        return self._pspeak.MaximumOfDerivativePos

    @property
    def maximum_of_derivative_sum(self) -> float:
        """Sum of the absolute values for both the positive and negative maximum derivative."""
        return self._pspeak.MaximumOfDerivativeSum

    @property
    def notes(self) -> str:
        """User notes stored on this peak."""
        return self._pspeak.Notes

    @property
    def y_offset(self) -> float:
        """Offset of Y."""
        return self._pspeak.OffsetY

    @property
    def index(self) -> int:
        """Location of the peak as index number of the curve."""
        return self._pspeak.PeakIndex

    @property
    def type(self) -> str:
        """Used to determine if a peak is auto found."""
        return str(self._pspeak.PeakType)

    @property
    def value(self) -> float:
        """Value of the peak in units of the curve.
        This is the value of the peak height relative to the baseline of the peak."""
        return self._pspeak.PeakValue

    @property
    def x(self) -> float:
        """X value of the peak."""
        return self._pspeak.PeakX

    @property
    def y(self) -> float:
        """Y value of the peak."""
        return self._pspeak.PeakY

    @property
    def right_index(self) -> int:
        """Left side of the peaks baseline as index number of the curve."""
        return self._pspeak.RightIndex

    @property
    def right_x(self) -> float:
        """X of the right side of the peak baseline."""
        return self._pspeak.RightX

    @property
    def right_y(self) -> float:
        """Returns the Y of the right side of the peak baseline."""
        return self._pspeak.RightY

    @property
    def width(self) -> float:
        """Full width at half-height of the peak."""
        return self._pspeak.Width
